from unittest import TestCase
import os
import h5py
import numpy as np
import toml
import orcasong.tools.make_data_split as mds

__author__ = 'Daniel Guderian'

test_dir = os.path.dirname(os.path.realpath(__file__))
test_data_dir = os.path.join(test_dir, "data")
#these are files that were processed with orcasong
mupage_file = os.path.join(test_data_dir, "processed_data_muon", "processed_graph_muon.h5")
neutrino_file = os.path.join(test_data_dir,"processed_data_neutrino", "processed_graph_neutrino.h5")
#config file containing 2 input groups
config_file = os.path.join(test_data_dir, "test_make_data_split_config.toml")
#the list files that will be created
list_file_dir = os.path.join(test_data_dir, "data_split_test_output", "conc_list_files")
list_output_val = os.path.join(list_file_dir, "test_list_validate_0.txt")
list_output_train = os.path.join("data_split_test_output", "conc_list_files", "test_list_train_0.txt")
#the scripts outputs
scripts_output_dir = os.path.join(test_data_dir, "data_split_test_output", "job_scripts")
concatenate_bash_script_train = os.path.join(scripts_output_dir, "concatenate_h5_test_list_train_0.sh")
concatenate_bash_script_val = os.path.join(scripts_output_dir, "concatenate_h5_test_list_validate_0.sh")
shuffle_bash_script_train = os.path.join(scripts_output_dir, "shuffle_h5_test_list_train_0.sh")
shuffle_bash_script_val = os.path.join(scripts_output_dir, "shuffle_h5_test_list_validate_0.sh")
#and the files that will be created from these scripts
concatenate_file = os.path.join("data_split_test_output", "data_split", "test_list_train_0.h5")

class TestMakeDataSplit(TestCase):
	
	''' Runs the make_data_split like in the use case. At the end, the created lists are checked.'''
	
	@classmethod
	def setUpClass(cls):
		#the expected lists to compare to
		cls.input_categories_list = ["neutrino","muon"]
		#include name with linebreak as they will look like this in the final files
		cls.file_path_list = ['processed_data_muon/processed_graph_muon.h5','processed_data_neutrino/processed_graph_neutrino.h5',
							  'processed_data_muon/processed_graph_muon.h5\n','processed_data_neutrino/processed_graph_neutrino.h5\n']
		cls.file_path_list_val = ['processed_data_neutrino/processed_graph_neutrino.h5','processed_data_neutrino/processed_graph_neutrino.h5\n']
		cls.n_events_list = [18,3]
		cls.contents_concatenate_script = ['concatenate ' + list_output_train + ' --outfile ' + concatenate_file]
		cls.contents_shuffle_script = ['h5shuffle2 ' + concatenate_file + ' --max_ram 1000000000 \n']

		
		#create list_file_dir
		if not os.path.exists(list_file_dir):
			os.makedirs(list_file_dir)
	
	@classmethod
	def tearDownClass(cls):
		#remove the lists created
		os.remove(list_output_val)
		os.remove(list_output_train)
		os.remove(concatenate_bash_script_train)
		os.remove(concatenate_bash_script_val)
		os.remove(shuffle_bash_script_train)
		os.remove(shuffle_bash_script_val)
		os.removedirs(scripts_output_dir)
		os.removedirs(list_file_dir)
		os.removedirs(os.path.join(test_data_dir, "data_split_test_output", "logs"))
		os.removedirs(os.path.join(test_data_dir, "data_split_test_output", "data_split"))

		

	def test_read_keys_off_config(self):
		self.cfg = read_config(config_file)
		#get input groups and compare
		self.ip_group_keys = mds.get_all_ip_group_keys(self.cfg)
		self.assertSequenceEqual(self.ip_group_keys,self.input_categories_list)
		
	def test_get_filepath_and_n_events(self):	
		#repeat first 2 steps
		self.cfg = read_config(config_file)
		self.ip_group_keys = mds.get_all_ip_group_keys(self.cfg)
		
		self.cfg,self.n_evts_total = update_cfg(self.cfg)
		
		for key in self.ip_group_keys:
			self.assertIn(self.cfg[key]['fpaths'][0],self.file_path_list)
			self.assertIn(self.cfg[key]['n_evts'],self.n_events_list)
	
	def test_make_split(self):
		#main
		#repeat first 3 steps
		self.cfg = read_config(config_file)
		self.ip_group_keys = mds.get_all_ip_group_keys(self.cfg)
		self.cfg,self.n_evts_total = update_cfg(self.cfg)
		
		self.cfg['n_evts_total'] = self.n_evts_total
		mds.print_input_statistics(self.cfg, self.ip_group_keys)
		for key in self.ip_group_keys:
			mds.add_fpaths_for_data_split_to_cfg(self.cfg, key)
		mds.make_dsplit_list_files(self.cfg)
		
		#assert the single output lists
		assert os.path.exists(list_output_val) == 1
		with open(list_output_val) as f:
			for line in f:
				self.assertIn(line,self.file_path_list_val)
		f.close
		
		assert os.path.exists(list_output_train) == 1
		with open(list_output_train) as f2:
			for line in f2:
				self.assertIn(line,self.file_path_list)
		f2.close
		
	def test_make_concatenate_and_shuffle_scripts(self):
		#main
		#repeat first 4 steps
		self.cfg = read_config(config_file)
		self.ip_group_keys = mds.get_all_ip_group_keys(self.cfg)
		self.cfg,self.n_evts_total = update_cfg(self.cfg)
		
		self.cfg['n_evts_total'] = self.n_evts_total
		mds.print_input_statistics(self.cfg, self.ip_group_keys)
		for key in self.ip_group_keys:
			mds.add_fpaths_for_data_split_to_cfg(self.cfg, key)
		mds.make_dsplit_list_files(self.cfg)
			
		#create the bash job scripts and test their content		
		mds.make_concatenate_and_shuffle_scripts(self.cfg)
		
		assert os.path.exists(concatenate_bash_script_train) == 1
		with open(concatenate_bash_script_train) as f:
			for line in f:
				pass		#yay, awesome style! ^^
			last_line = line
			self.assertIn(last_line,self.contents_concatenate_script)
		f.close

		assert os.path.exists(shuffle_bash_script_train) == 1
		with open(shuffle_bash_script_train) as f2:
			for line in f2:
				pass
			last_line = line
			self.assertIn(last_line,self.contents_shuffle_script)
		f2.close

		
def update_cfg(cfg):
	
	''' Update the cfg with file paths and also return the total number of events'''
	 
	#get input groups and compare
	ip_group_keys = mds.get_all_ip_group_keys(cfg)
	os.chdir(test_data_dir)
	n_evts_total = 0
	for key in ip_group_keys:
		print('Collecting information from input group ' + key)
		cfg[key]['fpaths'] = mds.get_h5_filepaths(cfg[key]['dir'])
		cfg[key]['n_files'] = len(cfg[key]['fpaths'])
		cfg[key]['n_evts'], cfg[key]['n_evts_per_file_mean'], cfg[key]['run_ids'] = mds.get_number_of_evts_and_run_ids(cfg[key]['fpaths'], dataset_key='y')
		n_evts_total += cfg[key]['n_evts']
			
	return cfg,n_evts_total
	
def read_config(config_file):
	#decode config
	cfg = toml.load(config_file)
	cfg['toml_filename'] = config_file
	return cfg
		



import os
from unittest import TestCase
import tempfile
import numpy as np
import h5py
import orcasong.core
import orcasong.extractors as extractors


__author__ = "Daniel Guderian"


test_dir = os.path.dirname(os.path.realpath(__file__))
NEUTRINO_FILE = os.path.join(test_dir, "data", "mcv6_ORCA_test_rc.gsg_tauCClowE-CC_3-50GeV.km3sim.jterbr00008155.jorcarec.jsh.aanet.760.h5")
DET_FILE_NEUTRINO = os.path.join(test_dir, "data", "KM3NeT_00000049_20200707.detx")

NO_COMPLE_RECO_FILE = os.path.join(test_dir, "data", "arca_test_without_some_jmuon_recos.h5")
ARCA_DETX = os.path.join(test_dir, "data", "KM3NeT_-00000001_20171212.detx")

class TestStdRecoExtractor(TestCase):
    """ Assert that the neutrino info is extracted correctly. File has 18 events. """

    @classmethod
    def setUpClass(cls):
        #normal case, with complete recos
        cls.proc = orcasong.core.FileGraph(
            max_n_hits=3,
            time_window=[0, 50],
            hit_infos=["pos_z", "time", "channel_id"],
            extractor=extractors.get_neutrino_mc_info_extr(NEUTRINO_FILE),
            det_file=DET_FILE_NEUTRINO,
            add_t0=True,
            keep_event_info=True,
        )
        cls.tmpdir = tempfile.TemporaryDirectory()
        cls.outfile = os.path.join(cls.tmpdir.name, "binned.h5")
        cls.proc.run(infile=NEUTRINO_FILE, outfile=cls.outfile)
        cls.f = h5py.File(cls.outfile, "r")
        cls.reco_names_in_original_file = ["best_jshower","best_jmuon"]
        cls.reco_names = ["jshower","jmuon"]
        cls.quantities_to_test = ["dir_z","dir_x","pos_x"]
        cls.mc_quantities_to_test = ["dir_z","dir_x","vertex_pos_x"]
        
        #case where a few recos from jmuon are missing
        cls.proc = orcasong.core.FileGraph(
            max_n_hits=3,
            time_window=[0, 50],
            hit_infos=["pos_z", "time", "channel_id"],
            extractor=extractors.get_neutrino_mc_info_extr(NO_COMPLE_RECO_FILE),
            det_file=ARCA_DETX,
            add_t0=True,
            keep_event_info=True,
        )
        cls.outfile_arca = os.path.join(cls.tmpdir.name, "binned_arca.h5")
        cls.proc.run(infile=NO_COMPLE_RECO_FILE, outfile=cls.outfile_arca)
        cls.f_arca = h5py.File(cls.outfile_arca, "r")

    @classmethod
    def tearDownClass(cls):
        cls.f.close()
        cls.tmpdir.cleanup()

    def test_keys(self):
        self.assertSetEqual(
            set(self.f.keys()),
            {
                "_i_event_info",
                "_i_group_info",
                "_i_y",
                "event_info",
                "group_info",
                "x",
                "x_indices",
                "y",
            },
        )

    def test_y_contents(self):
    
        #the y from the processed file 
        y = self.f["y"][()]
        
        #the info from the original file
        orig_info = h5py.File(NEUTRINO_FILE,"r")
        
        orig_reco_info = orig_info["reco"]
        orig_mc_info = orig_info["mc_tracks"]
        
        #test a few reco parameter
        for i in range(len(self.reco_names_in_original_file)):
            for j in self.quantities_to_test:
                orig_reco_values = orig_reco_info[self.reco_names_in_original_file[i]][j]
            
                assert np.allclose(orig_reco_values,y[self.reco_names[i]+"_"+j])
        
        #test a few mc truth parameter
        for j in range(len(self.quantities_to_test)):
            orig_mc_values = orig_mc_info[self.quantities_to_test[j]][0] #only take fist, the primary
            
            assert np.allclose(orig_mc_values,y[self.mc_quantities_to_test[j]][0])
    
    def test_incomplete_reco_padding(self): 
        
        #make sure the x and y even in this case have the same length
        x = self.f_arca["x"]
        y = self.f_arca["y"]
        
        assert len(x) == len(y)
    
    
    
    
    
    
    


from unittest import TestCase
import os
import h5py
import numpy as np
import orcasong.tools.postproc as postproc
import orcasong.tools.shuffle2 as shuffle2

__author__ = 'Stefan Reck'

test_dir = os.path.dirname(os.path.realpath(__file__))
MUPAGE_FILE = os.path.join(test_dir, "data", "mupage.root.h5")


class TestPostproc(TestCase):
    def setUp(self):
        self.output_file = "temp_output.h5"

    def tearDown(self):
        if os.path.exists(self.output_file):
            os.remove(self.output_file)

    def test_shuffle(self):
        postproc.postproc_file(
            input_file=MUPAGE_FILE,
            output_file=self.output_file,
            shuffle=True,
            event_skipper=None,
            delete=False,
            seed=13,
        )

        with h5py.File(self.output_file, "r") as f:
            np.testing.assert_equal(f["event_info"]["event_id"], np.array([1, 0, 2]))
            self.assertTrue("origin" in f.attrs.keys())


class TestShuffleV2(TestCase):
    def setUp(self):
        self.temp_input = "temp_input.h5"
        self.temp_output = "temp_output.h5"

        self.x, self.y = _make_shuffle_dummy_file(self.temp_input)
        np.random.seed(42)
        shuffle2.h5shuffle2(
            input_file=self.temp_input,
            output_file=self.temp_output,
            datasets=("x", "y"),
            max_ram=400,  # -> 2 batches
        )

    def tearDown(self):
        for f in (self.temp_input, self.temp_output):
            if os.path.exists(f):
                os.remove(f)

    def test_shuffled_has_same_entries_as_input(self):
        with h5py.File(self.temp_output, "r") as f:
            x_s = f["x"][()]
            np.testing.assert_array_equal(
                self.x[:, 1:], x_s[:, 1:][np.argsort(x_s[:, 0])]
            )

    def test_all_shuffled_datasets_have_same_order(self):
        with h5py.File(self.temp_output, "r") as f:
            np.testing.assert_array_equal(
                f["x"][:, 0], f["y"][:, 0]
            )

    def test_seed_produces_this_shuffled_order(self):
        target_order = np.array(
            [12.,  5.,  0., 20.,  6.,  2., 19., 13.,  1.,  9.,  8.,  4.,  3.,
             11., 16., 15.,  7., 10., 17., 14., 21., 18.])
        with h5py.File(self.temp_output, "r") as f:
            np.testing.assert_array_equal(
                f["x"][:, 0], target_order
            )

    def test_run_3_iterations(self):
        # just check if it goes through without errors
        fname = "temp_output_triple.h5"
        try:
            shuffle2.h5shuffle2(
                input_file=self.temp_input,
                output_file=fname,
                datasets=("x", "y"),
                iterations=3,
            )
        finally:
            if os.path.exists(fname):
                os.remove(fname)


def _make_shuffle_dummy_file(filepath):
    x = np.random.rand(22, 2)
    x[:, 0] = np.arange(22)
    y = np.random.rand(22, 3)
    y[:, 0] = np.arange(22)
    with h5py.File(filepath, "w") as f:
        f.create_dataset("x", data=x, chunks=(5, 2))
        f.create_dataset("y", data=y, chunks=(5, 3))
    return x, y



"""
Functions that extract info from a blob for the mc_info / y datafield
in the h5 files.

These are made for the specific given runs. They might not be
applicable to other data, and could cause errors or produce unexpected
results when used on data other then the specified. Check for example the 
primary position in the mc_tracks.

"""

import warnings
import numpy as np
from km3pipe.io.hdf5 import HDF5Header
from h5py import File

__author__ = "Daniel Guderian"
	

def get_std_reco(blob,rec_types,rec_parameters_names):

	"""
	Function to extract std reco info. This implementation requires h5 files 
	to be processed with the option "--best_tracks" which adds the selection
	of best tracks for each reco type to the output using the km3io tools.

	Returns
	-------
	std_reco_info : dict
			Dict with the std reco info of the best tracks.

	"""
	#this dict will be filled up
	std_reco_info = {}

	#all known reco types to iterate over
	reco_type_dict = {
			"BestJmuon" : ("jmuon_","best_jmuon"),
			"BestJshower" : ("jshower_","best_jshower"),
			"BestDusjshower" : ("dusjshower_","best_dusjshower"),
			"BestAashower" : ("aashower_","best_aashower"),
			}

	for name_in_blob,(identifier,best_track_name) in reco_type_dict.items():
	
		#always write out something for the generally present rec types
		if best_track_name in rec_types:
			
			#specific names are with the prefix from the rec type
			specific_reco_names = np.core.defchararray.add(identifier,rec_parameters_names)
			
			#extract actually present info
			if name_in_blob in blob:
				
				#get the previously identified best track
				bt = blob[name_in_blob]

				#get all its values
				values = bt.item()
				values_list = list(values)
				#reco_names = bt.dtype.names #in case the fitinf and stuff will be tailored to the reco types
												#at some point, get the names directly like this

			#in case there is no reco for this event but the reco type was done in general
			else:
				
				#fill all values with nan's
				values_array = np.empty(len(specific_reco_names))
				values_array[:] = np.nan
				values_list = values_array.tolist()
			
			#create a dict out of them
			keys_list = list(specific_reco_names)
			
			zip_iterator = zip(keys_list, values_list)
			reco_dict = dict(zip_iterator)

			#add this dict to the complete std reco collection
			std_reco_info.update(reco_dict)
			
	return std_reco_info


def get_rec_types_in_file(file):
	
	"""
	Checks and returns which rec types are in the file and thus need to be present 
	in all best track and their fitinf information later.
	"""

	#the known rec types
	rec_type_names = ["best_jmuon","best_jshower","best_dusjshower","best_aashower"]

	#all reco related objects in the file
	reco_objects_in_file = file["reco"].keys()
	
	#check which ones are in there
	rec_types_in_file = []
	for rec_type in rec_type_names:
		if rec_type in reco_objects_in_file:
			rec_types_in_file.append(rec_type)
			
			#also get from here the list of dtype names that is share for all recos
			rec_parameters_names = file["reco"][rec_type].dtype.names

	return rec_types_in_file,rec_parameters_names
	
def get_real_data_info_extr(input_file):

    """
    Wrapper function that includes the actual mc_info_extr
    for real data. There are no n_gen like in the neutrino case.

    Parameters
    ----------
    input_file : km3net data file
            Can be online or offline format.

    Returns
    -------
    mc_info_extr : function
            The actual mc_info_extr function that holds the extractions.

    """

    # check if std reco is present
    f = File(input_file, "r")
    has_std_reco = "reco" in f.keys()
    
    if has_std_reco:
        #also check, which rec types are present
        rec_types,rec_parameters_names = get_rec_types_in_file(f)

    def mc_info_extr(blob):

        """
        Processes a blob and creates the y with mc_info and, if existing, std reco.

        For this real data case it is only general event info, like the id.

        Parameters
        ----------
        blob : dict
                The blob from the pipeline.

        Returns
        -------
        track : dict
                Containing all the specified info the y should have.

        """

        event_info = blob["EventInfo"][0]

        # add n_hits info for the cut
        n_hits = len(blob["Hits"])

        track = {
            "event_id": event_info.event_id,
            "run_id": event_info.run_id,
            "trigger_mask": event_info.trigger_mask,
            "n_hits": n_hits,
        }

        # get all the std reco info
        if has_std_reco:
            
            std_reco_info = get_std_reco(blob,rec_types,rec_parameters_names)

            track.update(std_reco_info)

        return track

    return mc_info_extr


def get_random_noise_mc_info_extr(input_file):

    """
    Wrapper function that includes the actual mc_info_extr
    for random noise simulations. There are no n_gen like in the neutrino case.

    Parameters
    ----------
    input_file : km3net data file
            Can be online or offline format.

    Returns
    -------
    mc_info_extr : function
            The actual mc_info_extr function that holds the extractions.

    """

    # check if std reco is present
    f = File(input_file, "r")
    has_std_reco = "reco" in f.keys()
    
    if has_std_reco:
        #also check, which rec types are present
        rec_types,rec_parameters_names = get_rec_types_in_file(f)
    
    def mc_info_extr(blob):

        """
        Processes a blob and creates the y with mc_info and, if existing, std reco.

        For this random noise case it is only general event info, like the id.

        Parameters
        ----------
        blob : dict
                The blob from the pipeline.

        Returns
        -------
        track : dict
                Containing all the specified info the y should have.

        """
        event_info = blob["EventInfo"]

        track = {
            "event_id": event_info.event_id[0],
            "run_id": event_info.run_id[0],
            "particle_type": 0,
        }

        # get all the std reco info
        if has_std_reco:

            std_reco_info = get_std_reco(blob,rec_types,rec_parameters_names)

            track.update(std_reco_info)

        return track

    return mc_info_extr

	        
def get_neutrino_mc_info_extr(input_file):

    """
    Wrapper function that includes the actual mc_info_extr
    for neutrino simulations. The n_gen parameter, needed for neutrino weighting
    is extracted from the header of the file.

    Parameters
    ----------
    input_file : km3net data file
            Can be online or offline format.

    Returns
    -------
    mc_info_extr : function
            The actual mc_info_extr function that holds the extractions.

    """

    # check if std reco is present
    f = File(input_file, "r")
    has_std_reco = "reco" in f.keys()
    
    if has_std_reco:
        #also check, which rec types are present
        rec_types,rec_parameters_names = get_rec_types_in_file(f)

    # get the n_gen
    header = HDF5Header.from_hdf5(input_file)
    n_gen = header.genvol.numberOfEvents
        
    def mc_info_extr(blob):

        """
        Processes a blob and creates the y with mc_info and, if existing, std reco.

        For this neutrino case it is the full mc info for the primary neutrino; there are the several "McTracks":
        check the simulation which index "p" the neutrino has.

        Parameters
        ----------
        blob : dict
                The blob from the pipeline.

        Returns
        -------
        track : dict
                Containing all the specified info the y should have.

        """

        # get general info about the event
        event_id = blob["EventInfo"].event_id[0]
        run_id = blob["EventInfo"].run_id[0]
        # weights for neutrino analysis
        weight_w1 = blob["EventInfo"].weight_w1[0]
        weight_w2 = blob["EventInfo"].weight_w2[0]
        weight_w3 = blob["EventInfo"].weight_w3[0]

        # first, look for the particular neutrino index of the production
        p = 0  # for ORCA4 (and probably subsequent productions)

        mc_track = blob["McTracks"][p]

        # some track mc truth info
        particle_type = mc_track.pdgid      #sometimes type, sometimes pdgid
        energy = mc_track.energy
        is_cc = mc_track.cc
        bjorkeny = mc_track.by
        dir_x, dir_y, dir_z = mc_track.dir_x, mc_track.dir_y, mc_track.dir_z
        time_interaction = (
            mc_track.time
        )  # actually always 0 for primary neutrino, measured in MC time
        vertex_pos_x, vertex_pos_y, vertex_pos_z = (
            mc_track.pos_x,
            mc_track.pos_y,
            mc_track.pos_z,
        )

        # add also the nhits info
        n_hits = len(blob["Hits"])

        track = {
            "event_id": event_id,
            "particle_type": particle_type,
            "energy": energy,
            "is_cc": is_cc,
            "bjorkeny": bjorkeny,
            "dir_x": dir_x,
            "dir_y": dir_y,
            "dir_z": dir_z,
            "time_interaction": time_interaction,
            "run_id": run_id,
            "vertex_pos_x": vertex_pos_x,
            "vertex_pos_y": vertex_pos_y,
            "vertex_pos_z": vertex_pos_z,
            "n_hits": n_hits,
            "weight_w1": weight_w1,
            "weight_w2": weight_w2,
            "weight_w3": weight_w3,
            "n_gen": n_gen,
        }

        # get all the std reco info
        if has_std_reco:
            
            std_reco_info = get_std_reco(blob,rec_types,rec_parameters_names)

            track.update(std_reco_info)

        return track

    return mc_info_extr

#function used by Stefan to identify which muons leave how many mc hits in the (active) detector.
def get_mchits_per_muon(blob, inactive_du=None):

	"""
	For each muon in McTracks, get the number of McHits.
	Parameters
	----------
	blob
		The blob.
	inactive_du : int, optional
		McHits in this DU will not be counted.

	Returns
	-------
	np.array
		n_mchits, len = number of muons

	"""
	ids = blob["McTracks"]["id"]
	
	# Origin of each mchit (as int) in the active line
	origin = blob["McHits"]["origin"]
	
	if inactive_du:
		# only hits in active line
		origin = origin[blob["McHits"]["du"] != inactive_du]
	
	# get how many mchits were produced per muon in the bundle
	origin_dict = dict(zip(*np.unique(origin, return_counts=True)))
	
	return np.array([origin_dict.get(i, 0) for i in ids])


def get_muon_mc_info_extr(input_file,prod_identifier=2,inactive_du=None):

    """
    Wrapper function that includes the actual mc_info_extr
    for atm. muon simulations. There are no n_gen like in the neutrino case.

    Parameters
    ----------
    input_file : km3net data file
            Can be online or offline format.
	prod_identifier : int
		Solotion for now: just give a 1 for km3sim and a 2 for JSerine production. 
		They have different simulated run times, shich are needed for the correct scaling.

    Returns
    -------
    mc_info_extr : function
            The actual mc_info_extr function that holds the extractions.

    """

    # check if std reco is present
    f = File(input_file, "r")
    has_std_reco = "reco" in f.keys()
    
    if has_std_reco:
        #also check, which rec types are present
        rec_types,rec_parameters_names = get_rec_types_in_file(f)
    
    # no n_gen here, but needed for concatenation
    n_gen = 1
    
    def mc_info_extr(blob):

        """
        Processes a blob and creates the y with mc_info and, if existing, std reco.

        For this atm. muon case it is the full mc info for the primary; there are the several "McTracks":
        check the simulation to understand what "p" you want. Muons come in bundles that have the same direction.
        For energy: sum of all muons in a bundle,
        for vertex: weighted (energy) mean of the individual vertices .

        Parameters
        ----------
        blob : dict
                The blob from the pipeline.

        Returns
        -------
        track : dict
                Containing all the specified info the y should have.

        """

        event_id = blob["EventInfo"].event_id[0]
        run_id = blob["EventInfo"].run_id[0]

        p = 0  # for ORCA4 (and probably subsequent productions)

        mc_track = blob["McTracks"][p]

        particle_type = mc_track.pdgid # assumed that this is the same for all muons in a bundle, new: pdgid, old: type
        is_cc = 0 #set to 0
        bjorkeny = 0 #set to zero
        
        time_interaction = mc_track.time  # same for all muons in a bundle

        # sum up the energy from all muons that have at least x mc hits
        n_hits_per_muon = get_mchits_per_muon(blob, inactive_du=inactive_du) #DU1 in ORCA4 is in the detx but not powered
        
        #dont consider muons with less than 10 mc hits
        suficient_hits_mask = n_hits_per_muon >= 15
        energy = np.sum(blob["McTracks"][suficient_hits_mask].energy)
        
        # all muons in a bundle are parallel, so just take dir of first muon
        dir_x, dir_y, dir_z = mc_track.dir_x, mc_track.dir_y, mc_track.dir_z

        # vertex is the weighted (energy) mean of the individual vertices
        vertex_pos_x = np.average(
            blob["McTracks"][p:].pos_x, weights=blob["McTracks"][p:].energy
        )
        vertex_pos_y = np.average(
            blob["McTracks"][p:].pos_y, weights=blob["McTracks"][p:].energy
        )
        vertex_pos_z = np.average(
            blob["McTracks"][p:].pos_z, weights=blob["McTracks"][p:].energy
        )

        # add also the nhits info
        n_hits = len(blob["Hits"])

        # this is only relevant for neutrinos, though dummy info is needed for the concatenation
        weight_w1 = 10
        weight_w2 = 10
        weight_w3 = 10

        track = {
            "event_id": event_id,
            "particle_type": particle_type,
            "energy": energy,
            "is_cc": is_cc,
            "bjorkeny": bjorkeny,
            "dir_x": dir_x,
            "dir_y": dir_y,
            "dir_z": dir_z,
            "time_interaction": time_interaction,
            "run_id": run_id,
            "vertex_pos_x": vertex_pos_x,
            "vertex_pos_y": vertex_pos_y,
            "vertex_pos_z": vertex_pos_z,
            "n_hits": n_hits,
            "weight_w1": weight_w1,
            "weight_w2": weight_w2,
            "weight_w3": weight_w3,
            "n_gen": n_gen,
            "prod_identifier": prod_identifier,
        }

        # get all the std reco info
        if has_std_reco:
            
            std_reco_info = get_std_reco(blob,rec_types,rec_parameters_names)

            track.update(std_reco_info)

        return track

    return mc_info_extr



import tempfile
from unittest import TestCase
import numpy as np
import h5py
import orcasong.tools.concatenate as conc

__author__ = 'Stefan Reck'


class TestFileConcatenator(TestCase):
    """
    Test concatenation on pre-generated h5 files. They are in tests/data.

    """
    @classmethod
    def setUpClass(cls):
        cls.dummy_file_1 = tempfile.NamedTemporaryFile()
        _create_dummy_file(
            cls.dummy_file_1, columns=10, val_array=1, val_recarray=(1, 3)
        )
        cls.dummy_file_2 = tempfile.NamedTemporaryFile()
        _create_dummy_file(
            cls.dummy_file_2, columns=15, val_array=2, val_recarray=(4, 5)
        )
        cls.dummy_files = (
            cls.dummy_file_1.name,
            cls.dummy_file_2.name,
        )
        cls.compt_opts = {
            'complib': 'gzip', 'complevel': 1, 'chunksize': 5, "shuffle": False
        }

    @classmethod
    def tearDownClass(cls):
        cls.dummy_file_1.close()
        cls.dummy_file_2.close()

    def test_from_list(self):
        with tempfile.NamedTemporaryFile("w+") as tf:
            tf.writelines([f + "\n" for f in self.dummy_files])
            tf.seek(0)
            fc = conc.FileConcatenator.from_list(tf.name)
            self.assertSequenceEqual(self.dummy_files, fc.input_files)

    def test_get_compopts(self):
        comptopts = conc.get_compopts(self.dummy_files[0])
        self.assertDictEqual(comptopts, self.compt_opts)

    def test_fc_get_comptopts(self):
        fc = conc.FileConcatenator(self.dummy_files)
        self.assertDictEqual(fc.comptopts, self.compt_opts)

    def test_fc_get_comptopts_updates(self):
        fc = conc.FileConcatenator(self.dummy_files, comptopts_update={'chunksize': 1})
        target_compt_opts = dict(self.compt_opts)
        target_compt_opts["chunksize"] = 1
        self.assertDictEqual(fc.comptopts, target_compt_opts)

    def test_get_cumu_rows(self):
        fc = conc.FileConcatenator(self.dummy_files)
        np.testing.assert_array_equal(fc.cumu_rows, [0, 10, 25])

    def test_concatenate_used_files(self):
        fc = conc.FileConcatenator(self.dummy_files)
        with tempfile.TemporaryFile() as tf:
            fc.concatenate(tf)
            with h5py.File(tf, "r") as f:
                self.assertSequenceEqual(
                    f["used_files"][()].tolist(),
                    [n.encode("ascii", "ignore") for n in self.dummy_files],
                )

    def test_concatenate_attrs(self):
        fc = conc.FileConcatenator(self.dummy_files)
        with tempfile.TemporaryFile() as tf:
            fc.concatenate(tf)
            with h5py.File(tf, "r") as f:
                target_attrs = dict(f.attrs)
                target_dset_attrs = dict(f["numpy_array"].attrs)
            with h5py.File(self.dummy_files[0], "r") as f:
                source_attrs = dict(f.attrs)
                source_dset_attrs = dict(f["numpy_array"].attrs)

            self.assertDictEqual(source_attrs, target_attrs)
            self.assertDictEqual(source_dset_attrs, target_dset_attrs)

    def test_concatenate_array(self):
        fc = conc.FileConcatenator(self.dummy_files)
        with tempfile.TemporaryFile() as tf:
            fc.concatenate(tf)
            with h5py.File(tf, "r") as f:
                target = np.ones((25, 7, 3))
                target[10:, :, :] = 2.
                np.testing.assert_array_equal(
                    target,
                    f["numpy_array"][()]
                )

    def test_concatenate_recarray_with_groupid(self):
        fc = conc.FileConcatenator(self.dummy_files)
        with tempfile.TemporaryFile() as tf:
            fc.concatenate(tf)
            with h5py.File(tf, "r") as f:
                target = np.array(
                    [(1, 3, 1)] * 25,
                    dtype=[('x', '<f8'), ('y', '<i8'), ("group_id", "<i8")]
                )
                target["x"][10:] = 4.
                target["y"][10:] = 5.
                target["group_id"] = np.arange(25)
                np.testing.assert_array_equal(
                    target,
                    f["rec_array"][()]
                )


def _create_dummy_file(filepath, columns=10, val_array=1, val_recarray=(1, 3)):
    """ Create a dummy h5 file with an array and a recarray in it. """
    with h5py.File(filepath, "w") as f:
        dset = f.create_dataset(
            "numpy_array",
            data=np.ones(shape=(columns, 7, 3))*val_array,
            chunks=(5, 7, 3),
            compression="gzip",
            compression_opts=1
        )
        dset.attrs.create("test_dset", "ok")

        rec_array = np.array(
            [val_recarray + (1, )] * columns,
            dtype=[('x', '<f8'), ('y', '<i8'), ("group_id", "<i8")]
        )
        rec_array["group_id"] = np.arange(columns)
        f.create_dataset(
            "rec_array",
            data=rec_array,
            chunks=(5,),
            compression="gzip",
            compression_opts=1
        )
        f.attrs.create("test_file", "ok")



#!/usr/bin/env python
# -*- coding: utf-8 -*-


__author__ = "Michael Moser, Daniel Guderian"

import os
import warnings
import toml
import argparse
import h5py
import random
import numpy as np


def get_parser():
    # TODO deprecated
    warnings.warn("make_data_split is deprecated and has been renamed to orcasong make_data_split")
    parser = argparse.ArgumentParser(
        description="Create datasets based on the run_id's."
        "Use the config to add input folder and set the ranges."
        "Outputs a list in an txt file that can be used to "
        "concatenate the files specfied"
    )
    parser.add_argument(
        "config", type=str, help="See example config for detailed information"
    )

    return parser


def add_parser(subparsers):
    parser = subparsers.add_parser(
        "make_data_split",
        description="Create datasets based on the run_id's."
        "Use the config to add input folder and set the ranges."
        "Outputs a list in an txt file that can be used to "
        "concatenate the files specfied"
    )
    parser.add_argument(
        "config", type=str, help="See example config for detailed information"
    )
    parser.set_defaults(func=make_split)


def get_all_ip_group_keys(cfg):
    """
    Gets the keys of all input groups in the config dict.

    The input groups are defined as the dict elements, where the values have the type of a dict.

    Parameters
    ----------
    cfg : dict
        Dict that contains all configuration options and additional information.

    Returns
    -------
    ip_group_keys : list
        List of the input_group keys.

    """
    ip_group_keys = []
    for key in cfg:
        if type(cfg[key]) == dict:
            ip_group_keys.append(key)

    return ip_group_keys


def get_h5_filepaths(dirpath):
    """
    Returns the filepaths of all .h5 files that are located in a specific directory.

    Parameters
    ----------
    dirpath: str
        Path of the directory where the .h5 files are located.

    Returns
    -------
    filepaths : list
        List with the full filepaths of all .h5 files in the dirpath folder.

    """
    filepaths = []
    for f in os.listdir(dirpath):
        if f.endswith(".h5"):
            filepaths.append(dirpath + "/" + f)

    # randomize order
    random.Random(42).shuffle(filepaths)

    return filepaths


def get_number_of_evts(file, dataset_key="y"):
    """
    Returns the number of events of a file looking at the given dataset.

    Parameters
    ----------
    file : h5 file
        File to read the number of events from.
    dataset_key : str
        String which specifies, which dataset in a h5 file should be used for calculating the number of events.

    Returns
    -------
    n_evts : int
        The number of events in that file.

    """

    f = h5py.File(file, "r")
    dset = f[dataset_key]
    n_evts = dset.shape[0]
    f.close()

    return n_evts


def get_number_of_evts_and_run_ids(
    list_of_files, dataset_key="y", run_id_col_name="run_id"
):
    """
    Gets the number of events and the run_ids for all hdf5 files in the list_of_files.

    The number of events is calculated based on the dataset, which is specified with the dataset_key parameter.

    Parameters
    ----------
    list_of_files : list
        List which contains filepaths to h5 files.
    dataset_key : str
        String which specifies, which dataset in a h5 file should be used for calculating the number of events.
    run_id_col_name : str
        String, which specifies the column name of the 'run_id' column.

    Returns
    -------
    total_number_of_evts : int
        The cumulative (total) number of events.
    mean_number_of_evts_per_file : float
        The mean number of evts per file.
    run_ids : list
        List containing the run_ids of the files in the list_of_files.

    """

    total_number_of_evts = 0
    run_ids = []

    for i, fpath in enumerate(list_of_files):
        f = h5py.File(fpath, "r")

        dset = f[dataset_key]
        n_evts = dset.shape[0]
        total_number_of_evts += n_evts

        run_id = f[dataset_key][0][run_id_col_name]
        run_ids.append(run_id)

        f.close()

    mean_number_of_evts_per_file = total_number_of_evts / len(list_of_files)

    return total_number_of_evts, mean_number_of_evts_per_file, run_ids


def split(a, n):
    """
    Splits a list into n equal sized (if possible! if not, approximately) chunks.

    Parameters
    ----------
    a : list
        A list that should be split.
    n : int
        Number of times the input list should be split.

    Returns
    -------
    a_split : list
        The input list a, which has been split into n chunks.

    """
    # from https://stackoverflow.com/questions/2130016/splitting-a-list-into-n-parts-of-approximately-equal-length
    k, m = divmod(len(a), n)
    a_split = list(
        (a[i * k + min(i, m) : (i + 1) * k + min(i + 1, m)] for i in range(n))
    )
    return a_split


def print_input_statistics(cfg, ip_group_keys):
    """
    Prints some useful information for each input_group.

    Parameters
    ----------
    cfg : dict
        Dict that contains all configuration options and additional information.
    ip_group_keys : list
        List of the input_group keys.

    """

    print("----------------------------------------------------------------------")
    print("Printing input statistics for your " + cfg["toml_filename"] + " input:")
    print("----------------------------------------------------------------------")

    print(
        "Your input .toml file has the following data input groups: "
        + str(ip_group_keys)
    )
    print("Total number of events: " + str(cfg["n_evts_total"]))

    for key in ip_group_keys:
        print("--------------------------------------------------------------------")
        print("Info for group " + key + ":")
        print("Directory: " + cfg[key]["dir"])
        print("Total number of files: " + str(cfg[key]["n_files"]))
        print("Total number of events: " + str(cfg[key]["n_evts"]))
        print(
            "Mean number of events per file: "
            + str(round(cfg[key]["n_evts_per_file_mean"], 3))
        )
        print("--------------------------------------------------------------------")


def add_fpaths_for_data_split_to_cfg(cfg, key):
    """
    Adds all the filepaths for the output files into a list, and puts them into the cfg['output_dsplit'][key] location
    for all dsplits (train, validate, rest).

    Parameters
    ----------
    cfg : dict
        Dict that contains all configuration options and additional information.
    key : str
        The key of an input_group.

    """

    fpath_lists = {"train": [], "validate": [], "rest": []}
    for i, fpath in enumerate(cfg[key]["fpaths"]):

        run_id = cfg[key]["run_ids"][i]

        for dsplit in ["train", "validate", "rest"]:
            if "run_ids_" + dsplit in cfg[key]:
                if (
                    cfg[key]["run_ids_" + dsplit][0]
                    <= run_id
                    <= cfg[key]["run_ids_" + dsplit][1]
                ):
                    fpath_lists[dsplit].append(fpath)

    for dsplit in ["train", "validate", "rest"]:
        if len(fpath_lists[dsplit]) == 0:
            continue

        n_files_dsplit = cfg["n_files_" + dsplit]
        fpath_lists[dsplit] = split(fpath_lists[dsplit], n_files_dsplit)
        if "output_" + dsplit not in cfg:
            cfg["output_" + dsplit] = dict()
        cfg["output_" + dsplit][key] = fpath_lists[dsplit]


def make_dsplit_list_files(cfg):
    """
    Writes .list files of the datasplits to the disk, with the information in the cfg['output_dsplit'] dict.

    Parameters
    ----------
    cfg : dict
        Dict that contains all configuration options and additional information.

    """
    # check if //conc_list_files folder exists, if not create it.
    if not os.path.exists(cfg["output_file_folder"] + "/conc_list_files"):
        os.makedirs(cfg["output_file_folder"] + "/conc_list_files")

    print()
    print()
    print("In an run-by-run MC the run_id's might not be continuous.")
    print("Here are the actual numbers in the split sets:")
    print("----------------------------------------------")

    # loop over the different specified sets
    for dsplit in ["train", "validate", "rest"]:

        if "output_" + dsplit not in cfg:
            continue

        print(dsplit, "set:")

        first_key = list(cfg["output_" + dsplit].keys())[0]
        n_output_files = len(cfg["output_" + dsplit][first_key])

        # initialize counter of events for all input groups
        imput_groups_dict = cfg["output_" + dsplit]
        final_number_of_events = np.zeros(len(imput_groups_dict))

        # loop over the number of outputfiles for each set
        for i in range(n_output_files):
            fpath_output = (
                cfg["output_file_folder"]
                + "/conc_list_files/"
                + cfg["output_file_name"]
                + "_"
                + dsplit
                + "_"
                + str(i)
                + ".txt"
            )

            # save the txt list
            if "output_lists" not in cfg:
                cfg["output_lists"] = list()
            cfg["output_lists"].append(fpath_output)

            with open(fpath_output, "w") as f_out:
                for j in range(len(imput_groups_dict)):
                    keys = list(imput_groups_dict.keys())

                    for fpath in imput_groups_dict[keys[j]][i]:
                        # also count here the actual sizes
                        final_number_of_events[j] += get_number_of_evts(fpath)
                        f_out.write(fpath + "\n")

        # and then print them
        for i in range(len(imput_groups_dict)):
            print(keys[i], ":", int(final_number_of_events[i]))

        print("----------------------------------------------")


def make_concatenate_and_shuffle_scripts(cfg):
    """
    Function that writes qsub .sh files which concatenates all files inside the list files.

    Parameters
    ----------
    cfg : dict
        Dict that contains all configuration options and additional information.

    """

    dirpath = cfg["output_file_folder"]

    if not os.path.exists(
        dirpath + "/logs"
    ):  # check if /logs folder exists, if not create it.
        os.makedirs(dirpath + "/logs")
    if not os.path.exists(
        dirpath + "/job_scripts"
    ):  # check if /job_scripts folder exists, if not create it.
        os.makedirs(dirpath + "/job_scripts")
    if not os.path.exists(
        dirpath + "/data_split"
    ):  # check if /data_split folder exists, if not create it.
        os.makedirs(dirpath + "/data_split")

    # not available atm...
    # chunksize = '' if cfg['chunksize'] is None else ' --chunksize ' + str(cfg['chunksize'])
    # complib = '' if cfg['complib'] is None else ' --complib ' + str(cfg['complib'])
    # complevel = '' if cfg['complevel'] is None else ' --complevel ' + str(cfg['complevel'])

    # make qsub .sh file for concatenating
    for listfile_fpath in cfg["output_lists"]:
        listfile_fname = os.path.basename(listfile_fpath)
        listfile_fname_wout_ext = os.path.splitext(listfile_fname)[0]
        conc_outputfile_fpath = (
            cfg["output_file_folder"] + "/data_split/" + listfile_fname_wout_ext + ".h5"
        )

        fpath_bash_script = (
            dirpath + "/job_scripts/concatenate_h5_" + listfile_fname_wout_ext + ".sh"
        )

        with open(fpath_bash_script, "w") as f:
            f.write("#!/usr/bin/env bash\n")
            f.write("\n")
            f.write("source " + cfg["venv_path"] + "activate" + "\n")
            f.write("\n")
            f.write("# Concatenate the files in the list\n")

            f.write(
                "concatenate " + listfile_fpath + " --outfile " + conc_outputfile_fpath
            )
            # at the moment it is not possible to set the comp opts like this+ chunksize + complib + complevel

    # make qsub .sh file for shuffling

    for listfile_fpath in cfg["output_lists"]:
        listfile_fname = os.path.basename(listfile_fpath)
        listfile_fname_wout_ext = os.path.splitext(listfile_fname)[0]

        # This is the input for the shuffle tool!
        conc_outputfile_fpath = (
            cfg["output_file_folder"] + "/data_split/" + listfile_fname_wout_ext + ".h5"
        )

        fpath_bash_script = (
            dirpath + "/job_scripts/shuffle_h5_" + listfile_fname_wout_ext + ".sh"
        )

        with open(fpath_bash_script, "w") as f:
            f.write("#!/usr/bin/env bash\n")
            f.write("\n")
            f.write("source " + cfg["venv_path"] + "activate \n")
            f.write("\n")
            f.write("# Shuffle the h5 file \n")

            f.write(
                "h5shuffle2 " + conc_outputfile_fpath + " --max_ram 1000000000 \n"
            )  # fix to 1GB ram; in lyon using a fraction
            # is difficult...
            # time python shuffle/shuffle_h5.py'
            # + delete_flag_shuffle_tool
            # + chunksize + complib + complevel

            if cfg["shuffle_delete"]:
                f.write("\n")
                f.write("rm " + conc_outputfile_fpath + "\n")


def main():
    """
    Main function to make the data split.
    """

    # load the config
    parser = get_parser()
    parsed_args = parser.parse_args()
    make_split(parsed_args.config)


def make_split(config_file):
    # decode config
    cfg = toml.load(config_file)
    cfg["toml_filename"] = config_file

    # set some defaults/Nones - at the moment setting of the com opts is not available!
    # if 'chunksize' not in cfg: cfg['chunksize'] = None
    # if 'complib' not in cfg: cfg['complib'] = None
    # if 'complevel' not in cfg: cfg['complevel'] = None

    # read out all the input groups
    ip_group_keys = get_all_ip_group_keys(cfg)

    # and now loop over input groups extracting info
    n_evts_total = 0
    for key in ip_group_keys:
        print("Collecting information from input group " + key)
        cfg[key]["fpaths"] = get_h5_filepaths(cfg[key]["dir"])
        cfg[key]["n_files"] = len(cfg[key]["fpaths"])
        (
            cfg[key]["n_evts"],
            cfg[key]["n_evts_per_file_mean"],
            cfg[key]["run_ids"],
        ) = get_number_of_evts_and_run_ids(cfg[key]["fpaths"], dataset_key="y")

        n_evts_total += cfg[key]["n_evts"]

    cfg["n_evts_total"] = n_evts_total
    # print the extracted statistics
    print_input_statistics(cfg, ip_group_keys)

    if cfg["print_only"] is True:
        from sys import exit

        exit()

    for key in ip_group_keys:
        add_fpaths_for_data_split_to_cfg(cfg, key)

    # create the list files
    make_dsplit_list_files(cfg)

    # create bash scripts that can be submitted to do the concatenation and shuffle
    if cfg["make_qsub_bash_files"] is True:
        make_concatenate_and_shuffle_scripts(cfg)


if __name__ == "__main__":
    main()



"""
Scripts for postprocessing h5 files, e.g. shuffling.
"""
import os
import argparse
import warnings

import h5py
import km3pipe as kp
import km3modules as km
from orcasong.modules import EventSkipper
from orcasong.tools.concatenate import get_compopts, copy_attrs


def postproc_file(
        input_file,
        output_file=None,
        shuffle=True,
        event_skipper=None,
        delete=False,
        seed=42,
        statusbar_every=1000):
    """
    Postprocess a file using km3pipe after it has been preprocessed in OrcaSong.

    Parameters
    ----------
    input_file : str
        Path of the file that will be processed.
    output_file : str, optional
        If given, this will be the name of the output file.
        Otherwise, a name is auto generated.
    shuffle : bool
        Shuffle order of events.
    event_skipper : func, optional
        Function that takes the blob as an input, and returns a bool.
        If the bool is true, the event will be skipped.
    delete : bool
        Specifies if the input file should be deleted after processing.
    seed : int
        Sets a fixed random seed for the shuffling.
    statusbar_every : int or None
        After how many line a km3pipe status should be printed.

    Returns
    -------
    output_file : str
        Path to the output file.

    """
    if output_file is None:
        output_file = get_filepath_output(
            input_file, shuffle=shuffle, event_skipper=event_skipper)
    if os.path.exists(output_file):
        raise FileExistsError(output_file)

    print(f'Setting a Global Random State with the seed < {seed} >.')
    km.GlobalRandomState(seed=seed)

    comptopts = get_compopts(input_file)
    # km3pipe uses pytables for saving the shuffled output file,
    # which has the name 'zlib' for the 'gzip' filter
    if comptopts["complib"] == 'gzip':
        comptopts["complib"] = 'zlib'

    pipe = kp.Pipeline()
    if statusbar_every is not None:
        pipe.attach(km.common.StatusBar, every=statusbar_every)
        pipe.attach(km.common.MemoryObserver, every=statusbar_every)
    pipe.attach(
        kp.io.hdf5.HDF5Pump,
        filename=input_file,
        shuffle=shuffle,
        reset_index=True,
    )
    if event_skipper is not None:
        pipe.attach(EventSkipper, event_skipper=event_skipper)
    pipe.attach(
        kp.io.hdf5.HDF5Sink,
        filename=output_file,
        complib=comptopts["complib"],
        complevel=comptopts["complevel"],
        chunksize=comptopts["chunksize"],
        flush_frequency=1000,
    )
    pipe.drain()

    copy_used_files(input_file, output_file)
    copy_attrs(input_file, output_file)
    if delete:
        print("Deleting original file")
        os.remove(input_file)

    print("Done!")
    return output_file


def copy_used_files(source_file, target_file):
    """ Copy the "used_files" dataset from one h5 file to another, if it is present.
    """
    with h5py.File(source_file, "r") as src:
        if "used_files" in src:
            print("Copying used_files dataset")
            with h5py.File(target_file, "a") as trg:
                trg.create_dataset("used_files", data=src["used_files"])


def get_filepath_output(input_file, shuffle=True, event_skipper=None):
    """ Get the filename of the shuffled / rebalanced output file as a str.
    """
    fname_adtn = ''
    if shuffle:
        fname_adtn += '_shuffled'
    if event_skipper is not None:
        fname_adtn += '_reb'
    return f"{os.path.splitext(input_file)[0]}{fname_adtn}.h5"


def h5shuffle():
    # TODO deprecated
    warnings.warn("h5shuffle is deprecated and has been renamed to orcasong h5shuffle")
    parser = argparse.ArgumentParser(description='Shuffle an h5 file using km3pipe.')
    parser.add_argument('input_file', type=str, help='File to shuffle.')
    parser.add_argument('--output_file', type=str,
                        help='Name of output file. Default: Auto generate name.')
    parser.add_argument('--delete', action="store_true",
                        help='Delete original file afterwards.')

    postproc_file(**vars(parser.parse_args()), shuffle=True, event_skipper=None)



"""
Run OrcaSong functionalities from command line.

"""
import argparse
from orcasong.tools.concatenate import concatenate
from orcasong.tools.postproc import postproc_file
from orcasong.tools.shuffle2 import h5shuffle2
import orcasong.from_toml as from_toml
import orcasong.plotting.plot_binstats as plot_binstats
import orcasong.tools.make_data_split as make_data_split


def _add_parser_concatenate(subparsers):
    parser = subparsers.add_parser(
        "concatenate",
        description='Concatenate many small h5 files to a single large one '
                    'in a km3pipe compatible format. This is intended for '
                    'files that get generated by orcasong, i.e. all datsets '
                    'should have the same length, with one row per '
                    'blob. '
                    'Compression options and the datasets to be created in '
                    'the new file will be read from the first input file.')
    parser.add_argument(
        'file', type=str, nargs="*",
        help="Define the files to concatenate. If it's one argument: A txt list "
             "with pathes of h5 files to concatenate (one path per line). "
             "If it's multiple arguments: "
             "The pathes of h5 files to concatenate.")
    parser.add_argument(
        '--outfile', type=str, default="concatenated.h5",
        help='The absoulte filepath of the output .h5 file that will be created. ')
    parser.add_argument(
        '--no_used_files', action='store_true',
        help="Per default, the paths of the input files are added "
             "as their own datagroup in the output file. Use this flag to "
             "disable. ")
    parser.add_argument(
        '--skip_errors', action='store_true',
        help="If true, ignore files that can't be concatenated. ")
    parser.set_defaults(func=concatenate)


def _add_parser_h5shuffle(subparsers):
    parser = subparsers.add_parser(
        "h5shuffle",
        description='Shuffle an h5 file using km3pipe.',
    )
    parser.add_argument('input_file', type=str, help='File to shuffle.')
    parser.add_argument('--output_file', type=str,
                        help='Name of output file. Default: Auto generate name.')
    parser.add_argument('--delete', action="store_true",
                        help='Delete original file afterwards.')
    parser.set_defaults(func=postproc_file)


def _add_parser_h5shuffle2(subparsers):
    parser = subparsers.add_parser(
        "h5shuffle2",
        description="Shuffle datasets in a h5file that have the same length. "
        "Uses chunkwise readout for speed-up."
    )
    parser.add_argument(
        "input_file", type=str, help="Path of the file that will be shuffled."
    )
    parser.add_argument(
        "--output_file",
        type=str,
        default=None,
        help="If given, this will be the name of the output file. "
        "Default: input_file + suffix.",
    )
    parser.add_argument(
        "--datasets",
        type=str,
        nargs="*",
        default=("x", "y"),
        help="Which datasets to include in output. Default: x, y",
    )
    parser.add_argument(
        "--max_ram_fraction",
        type=float,
        default=0.25,
        help="in [0, 1]. Fraction of all available ram to use for reading one batch of data "
        "Note: this should "
        "be <=~0.25 or so, since lots of ram is needed for in-memory shuffling. "
        "Default: 0.25",
    )
    parser.add_argument(
        "--iterations",
        type=int,
        default=None,
        help="Shuffle the file this many times. Default: Auto choose best number.",
    )
    parser.add_argument(
        "--max_ram",
        type=int,
        default=None,
        help="Available ram in bytes. Default: Use fraction of maximum "
             "available instead (see max_ram_fraction).",
    )
    parser.set_defaults(func=h5shuffle2)


def _add_parser_version(subparsers):
    def show_version():
        from orcasong import version
        print(version)

    parser = subparsers.add_parser(
        "version",
        description="Show installed orcanet version.",
    )
    parser.set_defaults(func=show_version)


def main():
    parser = argparse.ArgumentParser(
        prog="orcasong",
        description=__doc__,
        formatter_class=argparse.RawTextHelpFormatter,
    )
    subparsers = parser.add_subparsers()

    from_toml.add_parser_run(subparsers)
    _add_parser_concatenate(subparsers)
    _add_parser_h5shuffle(subparsers)
    _add_parser_h5shuffle2(subparsers)
    plot_binstats.add_parser(subparsers)
    make_data_split.add_parser(subparsers)
    _add_parser_version(subparsers)

    kwargs = vars(parser.parse_args())
    func = kwargs.pop("func")
    func(**kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
import km3pipe as kp
import matplotlib.pyplot as plt

import orcasong.modules as modules

__author__ = 'Stefan Reck'


class TimePlotter:
    """
    For plotting the time distribution of hits, in a histogram,
    and finding a good binning.

    The plot will have some bins attached in both directions for
    better overview.

    Attributes:
    -----------
    files : list
        The .h5 files to read hits from.
    do_mc_hits : bool
        Read mchits instead of hits.
    det_file : str, optional
        Path to a .detx detector geometry file, which can be used to
        calibrate the hits.
    center_time : bool
        Subtract time of first triggered hit from all hit times. Will
        also be done for McHits if they are in the blob [default: True].
    add_t0 : bool
        If true, add t0 to the time of hits. If using a det_file,
        this will already have been done automatically [default: False].
    inactive_du : int, optional
        Dont plot hits from this du.

    """
    def __init__(self, files,
                 do_mchits=False,
                 det_file=None,
                 center_time=True,
                 add_t0=False,
                 subtract_t0_mchits=False,
                 inactive_du=None):
        if isinstance(files, str):
            self.files = [files]
        else:
            self.files = files
        self.do_mchits = do_mchits
        self.det_file = det_file
        self.add_t0 = add_t0
        self.subtract_t0_mchits = subtract_t0_mchits
        self.center_time = center_time
        self.inactive_du = inactive_du

        self.data = np.array([])
        for file in self.files:
            self._read(file, self.det_file)

    def hist(self, bins=50, padding=0.5, **kwargs):
        """
        Plot the hits as a histogram.

        Parameters
        ----------
        bins : int or np.array
            Number of bins, or bin edges. If bin edges are given, some
            bins are attached left and right (next to red vertical lines)
            for better overview.
        padding : float
            Fraction of total number of bins to attach left and right
            in the plot.

        """
        plt.grid(True, zorder=-10)
        plt.xlabel("time")
        if self.do_mchits:
            plt.ylabel("mchits / bin")
        else:
            plt.ylabel("hits / bin")
        if isinstance(bins, int):
            plt.hist(self.data, bins=bins, zorder=10, **kwargs)
        else:
            self.print_binstats(bins)
            plt.hist(
                self.data,
                bins=_get_padded_bins(bins, padding),
                zorder=10,
                **kwargs
            )
            for bin_line_x in (bins[0], bins[-1]):
                plt.axvline(
                    x=bin_line_x, color='firebrick', linestyle='--', zorder=20
                )

    def print_binstats(self, bin_edges):
        print(f"Cutoff left: {np.mean(self.data < bin_edges[0]):.2%}")
        print(f"Cutoff right: {np.mean(self.data > bin_edges[-1]):.2%}")
        print(f"Avg. time per bin: {np.mean(np.diff(bin_edges)):.2f}")

    def _read(self, file, det_file=None):
        if det_file is not None:
            cal = modules.DetApplier(det_file=det_file)
        else:
            cal = None
        if self.center_time or self.add_t0:
            time_pp = modules.TimePreproc(
                add_t0=self.add_t0,
                center_time=self.center_time,
                subtract_t0_mchits=self.subtract_t0_mchits,
            )
        else:
            time_pp = None

        pump = kp.io.hdf5.HDF5Pump(filename=file)
        for i, blob in enumerate(pump):
            if i % 1000 == 0:
                print("Blob {}".format(i))
            if cal is not None:
                blob = cal.process(blob)
            if time_pp is not None:
                blob = time_pp.process(blob)
            self.data = np.concatenate(
                [self.data, self._get_data_one_event(blob)])
        pump.close()

    def _get_data_one_event(self, blob):
        if self.do_mchits:
            fld = "McHits"
        else:
            fld = "Hits"
        blob_data = blob[fld]["time"]

        if self.inactive_du is not None:
            dus = blob[fld]["du"]
            blob_data = blob_data[dus != self.inactive_du]
        return blob_data


def _get_padded_bins(bin_edges, padding):
    """ Add fraction of total # of bins, with same width. """
    bin_width = bin_edges[1] - bin_edges[0]
    n_bins = len(bin_edges) - 1
    bins_to_add = np.ceil(padding * n_bins)
    return np.linspace(
        bin_edges[0] - bins_to_add * bin_width,
        bin_edges[-1] + bins_to_add * bin_width,
        int(n_bins + 2 * bins_to_add + 1),
    )



"""
Custom km3pipe modules for making nn input files.
"""

import numpy as np
import km3pipe as kp
import km3modules as km
import orcasong.plotting.plot_binstats as plot_binstats

__author__ = 'Stefan Reck'


class McInfoMaker(kp.Module):
    """
    Stores info as float64 in the blob.

    Attributes
    ----------
    extractor : function
        Function to extract the info. Takes the blob as input, outputs
        a dict with the desired mc_infos.
    store_as : str
        Store the mcinfo with this name in the blob.

    """

    def configure(self):
        self.extractor = self.require('extractor')
        self.store_as = self.require('store_as')
        self.to_float64 = self.get("to_float64", default=True)

    def process(self, blob):
        track = self.extractor(blob)
        if self.to_float64:
            dtypes = []
            for key, v in track.items():
                if key in ("group_id", "event_id"):
                    dtypes.append((key, type(v)))
                else:
                    dtypes.append((key, np.float64))
        else:
            dtypes = None
        kp_hist = kp.dataclasses.Table(
            track, dtype=dtypes,  h5loc='y', name='event_info')
        if len(kp_hist) != 1:
            self.log.warning(
                "Warning: Extracted mc_info should have len 1, "
                "but it has len {}".format(len(kp_hist))
            )
        blob[self.store_as] = kp_hist
        return blob


class TimePreproc(kp.Module):
    """
    Preprocess the time in the blob in various ways.

    Attributes
    ----------
    add_t0 : bool
        If true, t0 will be added to times of hits.
    center_time : bool
        If true, center hit and mchit times with the time of the first
        triggered hit.

    """

    def configure(self):
        self.add_t0 = self.get('add_t0', default=False)
        self.center_time = self.get('center_time', default=True)

        self._print_flags = set()

    def process(self, blob):
        if self.add_t0:
            blob = self.add_t0_time(blob)
        if self.center_time:
            blob = self.center_hittime(blob)
        return blob

    def add_t0_time(self, blob):
        self._print_once("Adding t0 to hit times")
        blob["Hits"].time = np.add(blob["Hits"].time, blob["Hits"].t0)
        return blob

    def center_hittime(self, blob):
        hits_time = blob["Hits"].time
        hits_triggered = blob["Hits"].triggered
        t_first_trigger = np.min(hits_time[hits_triggered != 0])

        self._print_once("Centering time of Hits with first triggered hit")
        blob["Hits"].time = np.subtract(hits_time, t_first_trigger)

        if "McHits" in blob:
            self._print_once("Centering time of McHits with first triggered hit")
            mchits_time = blob["McHits"].time
            blob["McHits"].time = np.subtract(mchits_time, t_first_trigger)

        return blob

    def _print_once(self, text):
        if text not in self._print_flags:
            self._print_flags.add(text)
            self.cprint(text)


class ImageMaker(kp.Module):
    """
    Make a n-d histogram from "Hits", and store it in the blob as 'samples'.

    Attributes
    ----------
    bin_edges_list : List
        List with the names of the fields to bin, and the respective bin edges,
        including the left- and right-most bin edge.
    hit_weights : str, optional
        Use blob["Hits"][hit_weights] as weights for samples in histogram.

    """

    def configure(self):
        self.bin_edges_list = self.require('bin_edges_list')
        self.hit_weights = self.get('hit_weights')
        self.store_as = "samples"

    def process(self, blob):
        data, bins, name = [], [], ""

        for bin_name, bin_edges in self.bin_edges_list:
            data.append(blob["Hits"][bin_name])
            bins.append(bin_edges)
            name += bin_name + "_"

        if self.hit_weights is not None:
            weights = blob["Hits"][self.hit_weights]
        else:
            weights = None

        histogram = np.histogramdd(data, bins=bins, weights=weights)[0]

        hist_one_event = histogram[np.newaxis, ...].astype(np.uint8)
        kp_hist = kp.dataclasses.NDArray(
            hist_one_event, h5loc='x', title=name + "event_images")

        blob[self.store_as] = kp_hist
        return blob


class BinningStatsMaker(kp.Module):
    """
    Generate a histogram of the number of hits for each binning field name.

    E.g. if the bin_edges_list contains "pos_z", this will make a histogram
    of #Hits vs. "pos_z", together with how many hits were outside
    of the bin edges in both directions.

    Per default, the resolution of the histogram (width of bins) will be
    higher then the given bin edges, and the edges will be stored seperatly.
    The time is the exception: The plotted bins have exactly the
    given bin edges.

    Attributes
    ----------
    bin_edges_list : List
        List with the names of the fields to bin, and the respective bin edges,
        including the left- and right-most bin edge.
    res_increase : int
        Increase the number of bins by this much in the hists (so that one
        can see if the edges have been placed correctly). Is never used
        for the time binning (field name "time").
    bin_plot_freq : int
        Extract data for the histograms only every given number of blobs
        (reduces time the pipeline takes to complete).

    """

    def configure(self):
        self.bin_edges_list = self.require('bin_edges_list')
        self.res_increase = self.get("res_increase", default=5)
        self.bin_plot_freq = 1

        self.hists = {}
        for bin_name, org_bin_edges in self.bin_edges_list:
            # dont space bin edges for time
            if bin_name == "time":
                bin_edges = org_bin_edges
            else:
                bin_edges = self._space_bin_edges(org_bin_edges)

            self.hists[bin_name] = {
                "hist": np.zeros(len(bin_edges) - 1),
                "hist_bin_edges": bin_edges,
                "bin_edges": org_bin_edges,
                # below smallest edge, above largest edge:
                "cut_off": np.zeros(2),
            }

        self.i = 0

    def _space_bin_edges(self, bin_edges):
        """
        Increase resolution of given binning.
        """
        increased_n_bins = (len(bin_edges) - 1) * self.res_increase + 1
        bin_edges = np.linspace(
            bin_edges[0], bin_edges[-1], increased_n_bins)

        return bin_edges

    def process(self, blob):
        """
        Extract data from blob for the hist plots.
        """
        if self.i % self.bin_plot_freq == 0:
            for bin_name, hists_data in self.hists.items():
                hist_bin_edges = hists_data["hist_bin_edges"]

                hits = blob["Hits"]
                data = hits[bin_name]
                # get how much is cut off due to these limits
                out_pos = data[data > np.max(hist_bin_edges)].size
                out_neg = data[data < np.min(hist_bin_edges)].size

                # get all hits which are not cut off by other bin edges
                data = hits[bin_name][self._is_in_limits(
                    hits, excluded=bin_name)]
                hist = np.histogram(data, bins=hist_bin_edges)[0]

                self.hists[bin_name]["hist"] += hist
                self.hists[bin_name]["cut_off"] += np.array([out_neg, out_pos])

        self.i += 1
        return blob

    def finish(self):
        """
        Append the hists, which are the stats of the binning.

        Its a dict with each binning field name containing the following
        ndarrays:

        bin_edges : The actual bin edges.
        cut_off : How many events were cut off in positive and negative
            direction due to this binning.
        hist_bin_edges : The bin edges for the plot in finer resolution then
            the actual bin edges.
        hist : The number of hist in each bin of the hist_bin_edges.

        """
        return self.hists

    def _is_in_limits(self, hits, excluded=None):
        """ Get which hits are in the limits defined by ALL bin edges
        (except for given one). """
        inside = None
        for dfield, edges in self.bin_edges_list:
            if dfield == excluded:
                continue
            is_in = np.logical_and(hits[dfield] >= min(edges),
                                   hits[dfield] <= max(edges))
            if inside is None:
                inside = is_in
            else:
                inside = np.logical_and(inside, is_in)
        return inside


class PointMaker(kp.Module):
    """
    Store individual hit info from "Hits" in the blob as 'samples'.

    Used for graph networks.

    Attributes
    ----------
    max_n_hits : int
        Maximum number of hits that gets saved per event. If an event has
        more, some will get cut!
    time_window : tuple, optional
        Two ints (start, end). Hits outside of this time window will be cut
        away (base on 'Hits/time').
        Default: Keep all hits.
    hit_infos : tuple, optional
        Which entries in the '/Hits' Table will be kept. E.g. pos_x, time, ...
        Default: Keep all entries.
    dset_n_hits : str, optional
        If given, store the number of hits that are in the time window
        as a new column called 'n_hits_intime' in the dataset with
        this name (usually this is EventInfo).

    """
    def configure(self):
        self.max_n_hits = self.require("max_n_hits")
        self.hit_infos = self.get("hit_infos", default=None)
        self.time_window = self.get("time_window", default=None)
        self.dset_n_hits = self.get("dset_n_hits", default=None)
        self.store_as = "samples"

    def process(self, blob):
        if self.hit_infos is None:
            self.hit_infos = blob["Hits"].dtype.names
        points, n_hits = self.get_points(blob)
        blob[self.store_as] = kp.NDArray(
            np.expand_dims(points, 0), h5loc="x", title="nodes")
        if self.dset_n_hits:
            blob[self.dset_n_hits] = blob[self.dset_n_hits].append_columns(
                "n_hits_intime", n_hits)
        return blob

    def get_points(self, blob):
        """
        Get the desired hit infos from the blob.

        Returns
        -------
        points : np.array
            The hit infos of this event as a 2d matrix. No of rows are
            fixed to the given max_n_hits. Each of the self.extract_keys,
            is in one column + an additional column which is 1 for
            actual hits, and 0 for if its a padded row.
        n_hits : int
            Number of hits in the given time window.

        """
        points = np.zeros(
            (self.max_n_hits, len(self.hit_infos) + 1), dtype="float32")

        hits = blob["Hits"]
        if self.time_window is not None:
            # remove hits outside of time window
            hits = hits[np.logical_and(
                hits["time"] >= self.time_window[0],
                hits["time"] <= self.time_window[1],
            )]

        n_hits = len(hits)
        if n_hits > self.max_n_hits:
            # if there are too many hits, take random ones, but keep order
            indices = np.arange(n_hits)
            np.random.shuffle(indices)
            which = indices[:self.max_n_hits]
            which.sort()
            hits = hits[which]

        for i, which in enumerate(self.hit_infos):
            data = hits[which]
            points[:n_hits, i] = data
        # last column is whether there was a hit or no
        points[:n_hits, -1] = 1.
        return points, n_hits

    def finish(self):
        return {"hit_infos": tuple(self.hit_infos) + ("is_valid", )}


class EventSkipper(kp.Module):
    """
    Skip events based on blob content.

    Attributes
    ----------
    event_skipper : callable
        Function that takes the blob as an input, and returns a bool.
        If the bool is true, the blob will be skipped.

    """

    def configure(self):
        self.event_skipper = self.require('event_skipper')
        self._not_skipped = 0
        self._skipped = 0

    def process(self, blob):
        if self.event_skipper(blob):
            self._skipped += 1
            return
        else:
            self._not_skipped += 1
            return blob

    def finish(self):
        tot_events = self._skipped + self._not_skipped
        self.cprint(
            f"Skipped {self._skipped}/{tot_events} events "
            f"({self._skipped/tot_events:.4%})."
        )


class DetApplier(kp.Module):
    """
    Apply calibration to the Hits and McHits with a detx file.

    Attributes
    ----------
    det_file : str
        Path to a .detx detector geometry file.
    correct_timeslew : bool
        If true, the time slewing of hits depending on their tot
        will be corrected.
    center_hits_to : tuple, optional
        Translate the xyz positions of the hits (and mchits), as if
        the detector was centered at the given position.
        E.g., if its (0, 0, None), the hits and mchits will be
        centered at xy = 00, and z will be left untouched.

    """

    def configure(self):
        self.det_file = self.require("det_file")
        self.correct_timeslew = self.get("correct_timeslew", default=True)
        self.center_hits_to = self.get("center_hits_to", default=None)

        self.cprint(f"Calibrating with {self.det_file}")
        self.calib = kp.calib.Calibration(filename=self.det_file)
        self._calib_checked = False

        # dict  dim_name: float
        self._vector_shift = None

        if self.center_hits_to:
            self._cache_shift_center()

    def process(self, blob):
        if self._calib_checked is False:
            if "pos_x" in blob["Hits"]:
                self.log.warn(
                    "Warning: Using a det file, but pos_x in Hits detected. "
                    "Is the file already calibrated? This might lead to "
                    "errors with t0."
                )
            self._calib_checked = True
        blob["Hits"] = self.calib.apply(
            blob["Hits"], correct_slewing=self.correct_timeslew)
        if "McHits" in blob:
            blob["McHits"] = self.calib.apply(blob["McHits"])
        if self.center_hits_to:
            self.shift_hits(blob)
        return blob

    def shift_hits(self, blob):
        """ Translate hits by cached vector. """
        for dim_name in ("pos_x", "pos_y", "pos_z"):
            blob["Hits"][dim_name] += self._vector_shift[dim_name]
            if "McHits" in blob:
                blob["McHits"][dim_name] += self._vector_shift[dim_name]

    def _cache_shift_center(self):
        det_center, shift = {}, {}
        for i, dim_name in enumerate(("pos_x", "pos_y", "pos_z")):
            center = self.calib.detector.dom_table[dim_name].mean()
            det_center[dim_name] = center

            if self.center_hits_to[i] is None:
                shift[dim_name] = 0
            else:
                shift[dim_name] = self.center_hits_to[i] - center

        self._vector_shift = shift
        self.cprint(f"original detector center: {det_center}")
        self.cprint(f"shift for hits: {self._vector_shift}")


class HitRotator(kp.Module):
    """
    Rotates hits by angle theta.

    Attributes
    ----------
    theta : float
        Angle by which hits are rotated (radian).

    """

    def configure(self):
        self.theta = self.require('theta')

    def process(self, blob):
        x = blob['Hits']['x']
        y = blob['Hits']['y']

        rot_matrix = np.array([[np.cos(self.theta), - np.sin(self.theta)],
                               [np.sin(self.theta), np.cos(self.theta)]])

        x_rot = []
        y_rot = []

        for i in range(0, len(x)):
            vec = np.array([[x[i]], [y[i]]])
            rot = np.dot(rot_matrix, vec)
            x_rot.append(rot[0][0])
            y_rot.append(rot[1][0])

        blob['Hits']['x'] = x_rot
        blob['Hits']['y'] = y_rot

        return blob




#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Code for computeing 2D/3D/4D histograms ("images") based on the event_hits hit pattern of the file_to_hits.py output"""

import km3pipe as kp
import matplotlib as mpl
mpl.use('Agg')
import matplotlib.pyplot as plt
import numpy as np
from mpl_toolkits.axes_grid1 import make_axes_locatable


def get_time_parameters(event_hits, mode=('trigger_cluster', 'all'), t_start_margin=0.15, t_end_margin=0.15):
    """
    Gets the fundamental time parameters in one place for cutting a time residual.

    Later on, these parameters cut out a certain time span during an event specified by t_start and t_end.

    Parameters
    ----------
    event_hits : ndarray(ndim=2)
        2D array that contains the hits data for a certain event_id.
    mode : tuple(str, str)
        Type of time cut that is used. Currently available: timeslice_relative and trigger_cluster.
    t_start_margin, t_end_margin : float
        Used in timeslice_relative mode. Defines the start/end time of the selected timespan with t_mean -/+ t_start * t_diff.

    Returns
    -------
    t_start, t_end : float
        Absolute start and end time that will be used for the later timespan cut.
        Events in this timespan are accepted, others are rejected.

    """
    t = event_hits[:, 3:4]

    if mode[0] == 'trigger_cluster':
        triggered = event_hits[:, 4:5]
        t = t[triggered == 1]
        t_mean = np.mean(t, dtype=np.float64)

        if mode[1] == 'tight-0':
            # make a tighter cut, 9.5ns / bin with 100 bins, useful for ORCA 115l mupage events
            t_start = t_mean - 450  # trigger-cluster - 350ns
            t_end = t_mean + 500  # trigger-cluster + 850ns

        elif mode[1] == 'tight-1':
            # make a tighter cut, 12.5ns / bin with 60 bins
            t_start = t_mean - 250  # trigger-cluster - 350ns
            t_end = t_mean + 500  # trigger-cluster + 850ns

        elif mode[1] == 'tight-2':
            # make an even tighter cut, 5.8ns / bin with 60 bins
            t_start = t_mean - 150  # trigger-cluster - 150ns
            t_end = t_mean + 200  # trigger-cluster + 200ns

        else:
            assert mode[1] == 'all' # 'all' refers to the common time range of neutrino events in 115l ORCA
            # include nearly all mc_hits from muon-CC and elec-CC, 20ns / bin with 60 bins
            t_start = t_mean - 350 # trigger-cluster - 350ns
            t_end = t_mean + 850 # trigger-cluster + 850ns

    elif mode[0] == 'timeslice_relative':
        t_min = np.amin(t)
        t_max = np.amax(t)
        t_diff = t_max - t_min
        t_mean = t_min + 0.5 * t_diff

        t_start = t_mean - t_start_margin * t_diff
        t_end = t_mean + t_end_margin * t_diff

    elif mode[0] is None:
        t_start = np.amin(t)
        t_end = np.amax(t)

    else: raise ValueError('Time cut modes other than "first_triggered" or "timeslice_relative" are currently not supported.')

    return t_start, t_end


def compute_4d_to_2d_histograms(event_hits, event_track, x_bin_edges, y_bin_edges, z_bin_edges, n_bins, timecut, do2d_plots, pdf_2d_plots):
    """
    Computes 2D numpy histogram 'images' from the 4D data and appends the 2D histograms to the all_4d_to_2d_hists list,
    [xy, xz, yz, xt, yt, zt].

    Parameters
    ----------
    event_hits : ndarray(ndim=2)
        2D array that contains the hits data for a certain event_id.
    event_track : ndarray(ndim=2)
        Contains the relevant mc_track info for the event in order to get a nice title for the pdf histos.
    x_bin_edges, y_bin_edges, z_bin_edges: ndarray(ndim=1)
        Bin edges for the X/Y/Z-direction.
    n_bins : tuple of int
        Contains the number of bins that should be used for each dimension.
    timecut : tuple(str, str/None)
        Tuple that defines what timecut should be used in hits_to_histograms.
    do2d_plots : bool
        If True, generate 2D matplotlib pdf histograms.
    pdf_2d_plots : mpl.backends.backend_pdf.PdfPages/None
        Either a mpl PdfPages instance or None.

    """
    x, y, z, t = event_hits[:, 0], event_hits[:, 1], event_hits[:, 2], event_hits[:, 3]

    # analyze time
    t_start, t_end = get_time_parameters(event_hits, mode=timecut)

    # create histograms for this event
    hist_xy = np.histogram2d(x, y, bins=(x_bin_edges, y_bin_edges))  # hist[0] = H, hist[1] = xedges, hist[2] = yedges
    hist_xz = np.histogram2d(x, z, bins=(x_bin_edges, z_bin_edges))
    hist_yz = np.histogram2d(y, z, bins=(y_bin_edges, z_bin_edges))

    hist_xt = np.histogram2d(x, t, bins=(x_bin_edges, n_bins[3]), range=((min(x_bin_edges), max(x_bin_edges)), (t_start, t_end)))
    hist_yt = np.histogram2d(y, t, bins=(y_bin_edges, n_bins[3]), range=((min(y_bin_edges), max(y_bin_edges)), (t_start, t_end)))
    hist_zt = np.histogram2d(z, t, bins=(z_bin_edges, n_bins[3]), range=((min(z_bin_edges), max(z_bin_edges)), (t_start, t_end)))

    if do2d_plots:
        hists = [hist_xy, hist_xz, hist_yz, hist_xt, hist_yt, hist_zt]
        convert_2d_numpy_hists_to_pdf_image(hists, t_start, t_end, pdf_2d_plots, event_track=event_track) # slow! takes about 1s per event

    hist_xy = kp.dataclasses.NDArray(hist_xy[0][np.newaxis, ...].astype(np.uint8), h5loc='x', title='XY_Event_Images')
    hist_xz = kp.dataclasses.NDArray(hist_xz[0][np.newaxis, ...].astype(np.uint8), h5loc='x', title='XZ_Event_Images')
    hist_yz = kp.dataclasses.NDArray(hist_yz[0][np.newaxis, ...].astype(np.uint8), h5loc='x', title='YZ_Event_Images')
    hist_xt = kp.dataclasses.NDArray(hist_xt[0][np.newaxis, ...].astype(np.uint8), h5loc='x', title='XT_Event_Images')
    hist_yt = kp.dataclasses.NDArray(hist_yt[0][np.newaxis, ...].astype(np.uint8), h5loc='x', title='YT_Event_Images')
    hist_zt = kp.dataclasses.NDArray(hist_zt[0][np.newaxis, ...].astype(np.uint8), h5loc='x', title='ZT_Event_Images')

    return hist_xy, hist_xz, hist_yz, hist_xt, hist_yt, hist_zt


def convert_2d_numpy_hists_to_pdf_image(hists, t_start, t_end, pdf_2d_plots, event_track=None):
    """
    Creates matplotlib 2D histos based on the numpy histogram2D objects and saves them to a pdf file.

    Parameters
    ----------
    hists : list(ndarray(ndim=2))
        Contains np.histogram2d objects of all projections [xy, xz, yz, xt, yt, zt].
    t_start, t_end : float
        Absolute start/end time of the timespan cut.
    pdf_2d_plots : mpl.backends.backend_pdf.PdfPages/None
        Either a mpl PdfPages instance or None.
    event_track : ndarray(ndim=2)
        Contains the relevant mc_track info for the event in order to get a nice title for the pdf histos.

    """

    fig = plt.figure(figsize=(10, 13))
    if event_track is not None:
        particle_type = {16: 'Tau', -16: 'Anti-Tau', 14: 'Muon', -14: 'Anti-Muon', 12: 'Electron', -12: 'Anti-Electron', 'isCC': ['NC', 'CC']}
        event_info = {'event_id': str(int(event_track.event_id[0])), 'energy': str(event_track.energy[0]),
                      'particle_type': particle_type[int(event_track.particle_type[0])],
                      'interaction_type': particle_type['isCC'][int(event_track.is_cc[0])]}
        title = event_info['particle_type'] + '-' + event_info['interaction_type'] + ', Event ID: ' + event_info['event_id'] + ', Energy: ' + event_info['energy'] + ' GeV'
        props = dict(boxstyle='round', facecolor='wheat', alpha=0.5)
        fig.suptitle(title, usetex=False, horizontalalignment='center', size='xx-large', bbox=props)

    t_diff = t_end - t_start

    axes_xy = plt.subplot2grid((3, 2), (0, 0), title='XY - projection', xlabel='X Position [m]', ylabel='Y Position [m]', aspect='equal', xlim=(-175, 175), ylim=(-175, 175))
    axes_xz = plt.subplot2grid((3, 2), (0, 1), title='XZ - projection', xlabel='X Position [m]', ylabel='Z Position [m]', aspect='equal', xlim=(-175, 175), ylim=(-57.8, 292.2))
    axes_yz = plt.subplot2grid((3, 2), (1, 0), title='YZ - projection', xlabel='Y Position [m]', ylabel='Z Position [m]', aspect='equal', xlim=(-175, 175), ylim=(-57.8, 292.2))

    axes_xt = plt.subplot2grid((3, 2), (1, 1), title='XT - projection', xlabel='X Position [m]', ylabel='Time [ns]', aspect='auto',
                               xlim=(-175, 175), ylim=(t_start - 0.1*t_diff, t_end + 0.1*t_diff))
    axes_yt = plt.subplot2grid((3, 2), (2, 0), title='YT - projection', xlabel='Y Position [m]', ylabel='Time [ns]', aspect='auto',
                               xlim=(-175, 175), ylim=(t_start - 0.1*t_diff, t_end + 0.1*t_diff))
    axes_zt = plt.subplot2grid((3, 2), (2, 1), title='ZT - projection', xlabel='Z Position [m]', ylabel='Time [ns]', aspect='auto',
                               xlim=(-57.8, 292.2), ylim=(t_start - 0.1*t_diff, t_end + 0.1*t_diff))

    def fill_subplot(hist_ab, axes_ab):
        # Mask hist_ab
        h_ab_masked = np.ma.masked_where(hist_ab[0] == 0, hist_ab[0])

        a, b = np.meshgrid(hist_ab[1], hist_ab[2]) #2,1

        # Format in classical numpy convention: x along first dim (vertical), y along second dim (horizontal)
        # Need to take that into account in convert_2d_numpy_hists_to_pdf_image()
        # transpose to get typical cartesian convention: y along first dim (vertical), x along second dim (horizontal)
        plot_ab = axes_ab.pcolormesh(a, b, h_ab_masked.T)

        the_divider = make_axes_locatable(axes_ab)
        color_axis = the_divider.append_axes("right", size="5%", pad=0.1)

        # add color bar
        cbar_ab = plt.colorbar(plot_ab, cax=color_axis, ax=axes_ab)
        cbar_ab.ax.set_ylabel('Hits [#]')

    fill_subplot(hists[0], axes_xy)
    fill_subplot(hists[1], axes_xz)
    fill_subplot(hists[2], axes_yz)
    fill_subplot(hists[3], axes_xt)
    fill_subplot(hists[4], axes_yt)
    fill_subplot(hists[5], axes_zt)

    fig.tight_layout(rect=[0, 0.02, 1, 0.95])
    pdf_2d_plots.savefig(fig)
    plt.close()


def compute_4d_to_3d_histograms(event_hits, x_bin_edges, y_bin_edges, z_bin_edges, n_bins, timecut):
    """
    Computes 3D numpy histogram 'images' from the 4D data and appends the 3D histograms to the all_4d_to_3d_hists list,
    [xyz, xyt, xzt, yzt, rzt].

    Careful: Currently, appending to all_4d_to_3d_hists takes quite a lot of memory (about 200MB for 3500 events).
    In the future, the list should be changed to a numpy ndarray.
    (Which unfortunately would make the code less readable, since an array is needed for each projection...)

    Parameters
    ----------
    event_hits : ndarray(ndim=2)
        2D array that contains the hits data for a certain event_id.
    x_bin_edges, y_bin_edges, z_bin_edges : ndarray(ndim=2)
        Bin edges for the X/Y/Z-direction.
    n_bins : tuple of int
        Contains the number of bins that should be used for each dimension.
    timecut : tuple(str, str/None)
        Tuple that defines what timecut should be used in hits_to_histograms.

    """

    x, y, z, t = event_hits[:, 0:1], event_hits[:, 1:2], event_hits[:, 2:3], event_hits[:, 3:4]

    t_start, t_end = get_time_parameters(event_hits, mode=timecut)

    hist_xyz = np.histogramdd(event_hits[:, 0:3], bins=(x_bin_edges, y_bin_edges, z_bin_edges))

    hist_xyt = np.histogramdd(np.concatenate([x, y, t], axis=1), bins=(x_bin_edges, y_bin_edges, n_bins[3]),
                              range=((min(x_bin_edges), max(x_bin_edges)), (min(y_bin_edges), max(y_bin_edges)), (t_start, t_end)))
    hist_xzt = np.histogramdd(np.concatenate([x, z, t], axis=1), bins=(x_bin_edges, z_bin_edges, n_bins[3]),
                              range=((min(x_bin_edges), max(x_bin_edges)), (min(z_bin_edges), max(z_bin_edges)), (t_start, t_end)))
    hist_yzt = np.histogramdd(event_hits[:, 1:4], bins=(y_bin_edges, z_bin_edges, n_bins[3]),
                              range=((min(y_bin_edges), max(y_bin_edges)), (min(z_bin_edges), max(z_bin_edges)), (t_start, t_end)))

    # add a rotation-symmetric 3d hist
    r = np.sqrt(x * x + y * y)
    rzt = np.concatenate([r, z, t], axis=1)
    hist_rzt = np.histogramdd(rzt, bins=(n_bins[0], n_bins[2], n_bins[3]), range=((np.amin(r), np.amax(r)), (np.amin(z), np.amax(z)), (t_start, t_end)))

    hist_xyz = kp.dataclasses.NDArray(hist_xyz[0][np.newaxis, ...].astype(np.uint8), h5loc='x', title='XYZ_Event_Images')
    hist_xyt = kp.dataclasses.NDArray(hist_xyt[0][np.newaxis, ...].astype(np.uint8), h5loc='x', title='XYT_Event_Images')
    hist_xzt = kp.dataclasses.NDArray(hist_xzt[0][np.newaxis, ...].astype(np.uint8), h5loc='x', title='XZT_Event_Images')
    hist_yzt = kp.dataclasses.NDArray(hist_yzt[0][np.newaxis, ...].astype(np.uint8), h5loc='x', title='YZT_Event_Images')
    hist_rzt = kp.dataclasses.NDArray(hist_rzt[0][np.newaxis, ...].astype(np.uint8), h5loc='x', title='RZT_Event_Images')

    return hist_xyz, hist_xyt, hist_xzt, hist_yzt, hist_rzt


def compute_4d_to_4d_histograms(event_hits, x_bin_edges, y_bin_edges, z_bin_edges, n_bins, timecut, do4d):
    """
    Computes 4D numpy histogram 'images' from the 4D data and appends the 4D histogram to the all_4d_to_4d_hists list,
    [xyzt / xyzc]

    Parameters
    ----------
    event_hits : ndarray(ndim=2)
        2D array that contains the hits data for a certain event_id.
    x_bin_edges, y_bin_edges, z_bin_edges : ndarray(ndim=2)
        Bin edges for the X/Y/Z-direction.
    n_bins : tuple of int
        Contains the number of bins that should be used for each dimension.
    timecut : tuple(str, str/None)
        Tuple that defines what timecut should be used in hits_to_histograms.
    do4d : tuple(bool, str)
        Tuple, where [1] declares what should be used as 4th dimension after xyz.
        Currently, only 'time' and 'channel_id' are available.

    """
    t_start, t_end = get_time_parameters(event_hits, mode=timecut)

    if do4d[1] == 'time':
        hist_4d = np.histogramdd(event_hits[:, 0:4], bins=(x_bin_edges, y_bin_edges, z_bin_edges, n_bins[3]),
                                   range=((min(x_bin_edges),max(x_bin_edges)),(min(y_bin_edges),max(y_bin_edges)),
                                          (min(z_bin_edges),max(z_bin_edges)),(t_start, t_end)))

    elif do4d[1] == 'channel_id':
        time = event_hits[:, 3]
        event_hits = event_hits[np.logical_and(time >= t_start, time <= t_end)]
        channel_id = event_hits[:, 5:6]
        hist_4d = np.histogramdd(np.concatenate([event_hits[:, 0:3], channel_id], axis=1), bins=(x_bin_edges, y_bin_edges, z_bin_edges, 31),
                                   range=((min(x_bin_edges),max(x_bin_edges)),(min(y_bin_edges),max(y_bin_edges)),
                                          (min(z_bin_edges),max(z_bin_edges)),(np.amin(channel_id), np.amax(channel_id))))

    else:
        raise ValueError('The parameter in do4d[1] ' + str(do4d[1]) + ' is not available. Currently, only time and channel_id are supported.')

    proj_name = 'XYZT' if not do4d[1] == 'channel_id' else 'XYZC'
    title = proj_name + '_Event_Images'
    hist_4d = kp.dataclasses.NDArray(hist_4d[0][np.newaxis, ...].astype(np.uint8), h5loc='x', title=title)

    return hist_4d


class HistogramMaker(kp.Module):
    """
    Class that takes a km3pipe blob which contains the information for one event and returns
    a blob with a hit array and a track array that contains all relevant information of the event.
    """
    def configure(self):
        """
        Sets up the input arguments of the EventDataExtractor class.
        """
        self.x_bin_edges = self.require('x_bin_edges')
        self.y_bin_edges = self.require('y_bin_edges')
        self.z_bin_edges = self.require('z_bin_edges')
        self.n_bins = self.require('n_bins')
        self.timecut = self.require('timecut')
        self.do2d = self.require('do2d')
        self.do2d_plots = self.require('do2d_plots')
        self.pdf_2d_plots = self.get('pdf_2d_plots')
        self.do3d = self.require('do3d')
        self.do4d = self.require('do4d')

        self.i = 0

    def process(self, blob):
        """
        Returns a blob (dict), which contains the event_hits array and the event_track array.

        Parameters
        ----------
        blob : dict
            Km3pipe blob which contains all the data from the input file.

        Returns
        -------
        blob : dict
            Dictionary that contains the event_hits array and the event_track array.

        """
        if self.do2d:
            blob['xy'], blob['xz'], blob['yz'], blob['xt'], blob['yt'], blob['zt'] = compute_4d_to_2d_histograms(
                blob['event_hits'], blob['event_track'], self.x_bin_edges, self.y_bin_edges, self.z_bin_edges,
                self.n_bins, self.timecut, self.do2d_plots[0], self.pdf_2d_plots)

            self.i += 1
            if self.pdf_2d_plots is not None and self.i >= self.do2d_plots[1]:
                self.pdf_2d_plots.close()
                raise StopIteration

        if self.do3d:
            blob['xyz'], blob['xyt'], blob['xzt'], blob['yzt'], blob['rzt'] = compute_4d_to_3d_histograms(
                blob['event_hits'], self.x_bin_edges, self.y_bin_edges, self.z_bin_edges,
                self.n_bins, self.timecut)

        if self.do4d[0]:
            proj_key = 'xyzt' if not self.do4d[1] == 'channel_id' else 'xyzc'

            blob[proj_key] = compute_4d_to_4d_histograms(blob['event_hits'], self.x_bin_edges, self.y_bin_edges, self.z_bin_edges,
                                        self.n_bins, self.timecut, self.do4d)

        return blob


import os
import time
import datetime
import argparse
import warnings

import numpy as np
import psutil
import h5py
from km3pipe.sys import peak_memory_usage

from orcasong.tools.postproc import get_filepath_output, copy_used_files
from orcasong.tools.concatenate import copy_attrs


__author__ = "Stefan Reck"


def h5shuffle2(
    input_file,
    output_file=None,
    iterations=None,
    datasets=("x", "y"),
    max_ram_fraction=0.25,
    max_ram=None,
):
    if output_file is None:
        output_file = get_filepath_output(input_file, shuffle=True)
    if iterations is None:
        iterations = get_n_iterations(
            input_file,
            datasets=datasets,
            max_ram_fraction=max_ram_fraction,
            max_ram=max_ram,
        )
    np.random.seed(42)
    for i in range(iterations):
        print(f"\nIteration {i+1}/{iterations}")
        if iterations == 1:
            # special case if theres only one iteration
            stgs = {
                "input_file": input_file,
                "output_file": output_file,
                "delete": False,
            }
        elif i == 0:
            # first iteration
            stgs = {
                "input_file": input_file,
                "output_file": f"{output_file}_temp_{i}",
                "delete": False,
            }
        elif i == iterations - 1:
            # last iteration
            stgs = {
                "input_file": f"{output_file}_temp_{i-1}",
                "output_file": output_file,
                "delete": True,
            }
        else:
            # intermediate iterations
            stgs = {
                "input_file": f"{output_file}_temp_{i-1}",
                "output_file": f"{output_file}_temp_{i}",
                "delete": True,
            }
        shuffle_file(
            datasets=datasets,
            max_ram=max_ram,
            max_ram_fraction=max_ram_fraction,
            chunks=True,
            **stgs,
        )


def shuffle_file(
    input_file,
    datasets=("x", "y"),
    output_file=None,
    max_ram=None,
    max_ram_fraction=0.25,
    chunks=False,
    delete=False,
):
    """
    Shuffle datasets in a h5file that have the same length.

    Parameters
    ----------
    input_file : str
        Path of the file that will be shuffle.
    datasets : tuple
        Which datasets to include in output.
    output_file : str, optional
        If given, this will be the name of the output file.
        Otherwise, a name is auto generated.
    max_ram : int, optional
        Available ram in bytes. Default: Use fraction of
        maximum available (see max_ram_fraction).
    max_ram_fraction : float
        in [0, 1]. Fraction of ram to use for reading one batch of data
        when max_ram is None. Note: when using chunks, this should
        be <=~0.25, since lots of ram is needed for in-memory shuffling.
    chunks : bool
        Use chunk-wise readout. Large speed boost, but will
        only quasi-randomize order! Needs lots of ram
        to be accurate! (use a node with at least 32gb, the more the better)
    delete : bool
        Delete the original file afterwards?

    Returns
    -------
    output_file : str
        Path to the output file.

    """
    start_time = time.time()
    if output_file is None:
        output_file = get_filepath_output(input_file, shuffle=True)
    if os.path.exists(output_file):
        raise FileExistsError(output_file)
    if max_ram is None:
        max_ram = get_max_ram(max_ram_fraction)

    temp_output_file = (
        output_file + "_temp_" + time.strftime("%d-%m-%Y-%H-%M-%S", time.gmtime())
    )
    with h5py.File(input_file, "r") as f_in:
        dset_infos, n_lines = get_dset_infos(f_in, datasets, max_ram)
        print(f"Shuffling datasets {datasets} with {n_lines} lines each")

        if not chunks:
            indices = np.arange(n_lines)
            np.random.shuffle(indices)

            with h5py.File(temp_output_file, "x") as f_out:
                for dset_info in dset_infos:
                    print("Creating dataset", dset_info["name"])
                    make_dset(f_out, dset_info, indices)
                    print("Done!")
        else:
            indices_chunked = get_indices_largest(dset_infos)

            with h5py.File(temp_output_file, "x") as f_out:
                for dset_info in dset_infos:
                    print("Creating dataset", dset_info["name"])
                    make_dset_chunked(f_out, dset_info, indices_chunked)
                    print("Done!")

    copy_used_files(input_file, temp_output_file)
    copy_attrs(input_file, temp_output_file)
    os.rename(temp_output_file, output_file)
    if delete:
        os.remove(input_file)
    print(
        f"Elapsed time: " f"{datetime.timedelta(seconds=int(time.time() - start_time))}"
    )
    return output_file


def get_max_ram(max_ram_fraction):
    max_ram = max_ram_fraction * psutil.virtual_memory().available
    print(f"Using {max_ram_fraction:.2%} of available ram = {max_ram} bytes")
    return max_ram


def get_indices_largest(dset_infos):
    largest_dset = np.argmax([v["n_batches_chunkwise"] for v in dset_infos])
    dset_info = dset_infos[largest_dset]

    print(f"Total chunks: {dset_info['n_chunks']}")
    ratio = dset_info["chunks_per_batch"] / dset_info["n_chunks"]
    print(f"Chunks per batch: {dset_info['chunks_per_batch']} ({ratio:.2%})")

    return get_indices_chunked(
        dset_info["n_batches_chunkwise"],
        dset_info["n_chunks"],
        dset_info["chunksize"],
    )


def get_n_iterations(
    input_file, datasets=("x", "y"), max_ram=None, max_ram_fraction=0.25
):
    """ Get how often you have to shuffle with given ram to get proper randomness. """
    if max_ram is None:
        max_ram = get_max_ram(max_ram_fraction=max_ram_fraction)
    with h5py.File(input_file, "r") as f_in:
        dset_infos, n_lines = get_dset_infos(f_in, datasets, max_ram)
    largest_dset = np.argmax([v["n_batches_chunkwise"] for v in dset_infos])
    dset_info = dset_infos[largest_dset]
    n_iterations = int(
        np.ceil(np.log(dset_info["n_chunks"]) / np.log(dset_info["chunks_per_batch"]))
    )
    print(f"Total chunks: {dset_info['n_chunks']}")
    print(f"Chunks per batch: {dset_info['chunks_per_batch']}")
    print(f"--> min iterations for full shuffle: {n_iterations}")
    return n_iterations


def get_indices_chunked(n_batches, n_chunks, chunksize):
    """ Return a list with the chunkwise shuffled indices of each batch. """
    chunk_indices = np.arange(n_chunks)
    np.random.shuffle(chunk_indices)
    chunk_batches = np.array_split(chunk_indices, n_batches)

    index_batches = []
    for bat in chunk_batches:
        idx = (bat[:, None] * chunksize + np.arange(chunksize)[None, :]).flatten()
        np.random.shuffle(idx)
        index_batches.append(idx)

    return index_batches


def get_dset_infos(f, datasets, max_ram):
    """ Check datasets and retrieve relevant infos for each. """
    dset_infos = []
    n_lines = None
    for i, name in enumerate(datasets):
        dset = f[name]
        if i == 0:
            n_lines = len(dset)
        else:
            if len(dset) != n_lines:
                raise ValueError(
                    f"dataset {name} has different length! " f"{len(dset)} vs {n_lines}"
                )
        chunksize = dset.chunks[0]
        n_chunks = int(np.ceil(n_lines / chunksize))
        # TODO in h5py 3.X, use .nbytes to get uncompressed size
        bytes_per_line = np.asarray(dset[0]).nbytes
        bytes_per_chunk = bytes_per_line * chunksize

        lines_per_batch = int(np.floor(max_ram / bytes_per_line))
        chunks_per_batch = int(np.floor(max_ram / bytes_per_chunk))

        dset_infos.append(
            {
                "name": name,
                "dset": dset,
                "chunksize": chunksize,
                "n_lines": n_lines,
                "n_chunks": n_chunks,
                "bytes_per_line": bytes_per_line,
                "bytes_per_chunk": bytes_per_chunk,
                "lines_per_batch": lines_per_batch,
                "chunks_per_batch": chunks_per_batch,
                "n_batches_linewise": int(np.ceil(n_lines / lines_per_batch)),
                "n_batches_chunkwise": int(np.ceil(n_chunks / chunks_per_batch)),
            }
        )
    return dset_infos, n_lines


def make_dset(f_out, dset_info, indices):
    """ Create a shuffled dataset in the output file. """
    for batch_index in range(dset_info["n_batches_linewise"]):
        print(f"Processing batch {batch_index+1}/{dset_info['n_batches_linewise']}")

        slc = slice(
            batch_index * dset_info["lines_per_batch"],
            (batch_index + 1) * dset_info["lines_per_batch"],
        )
        to_read = indices[slc]
        # reading has to be done with linearly increasing index,
        #  so sort -> read -> undo sorting
        sort_ix = np.argsort(to_read)
        unsort_ix = np.argsort(sort_ix)
        data = dset_info["dset"][to_read[sort_ix]][unsort_ix]

        if batch_index == 0:
            in_dset = dset_info["dset"]
            out_dset = f_out.create_dataset(
                dset_info["name"],
                data=data,
                maxshape=in_dset.shape,
                chunks=in_dset.chunks,
                compression=in_dset.compression,
                compression_opts=in_dset.compression_opts,
                shuffle=in_dset.shuffle,
            )
            out_dset.resize(len(in_dset), axis=0)
        else:
            f_out[dset_info["name"]][slc] = data


def make_dset_chunked(f_out, dset_info, indices_chunked):
    """ Create a shuffled dataset in the output file. """
    start_idx = 0
    for batch_index, to_read in enumerate(indices_chunked):
        print(f"Processing batch {batch_index+1}/{len(indices_chunked)}")

        # remove indices outside of dset
        to_read = to_read[to_read < len(dset_info["dset"])]

        # reading has to be done with linearly increasing index
        #  fancy indexing is super slow
        #  so sort -> turn to slices -> read -> conc -> undo sorting
        sort_ix = np.argsort(to_read)
        unsort_ix = np.argsort(sort_ix)
        fancy_indices = to_read[sort_ix]
        slices = slicify(fancy_indices)
        data = np.concatenate([dset_info["dset"][slc] for slc in slices])
        data = data[unsort_ix]

        if batch_index == 0:
            in_dset = dset_info["dset"]
            out_dset = f_out.create_dataset(
                dset_info["name"],
                data=data,
                maxshape=in_dset.shape,
                chunks=in_dset.chunks,
                compression=in_dset.compression,
                compression_opts=in_dset.compression_opts,
                shuffle=in_dset.shuffle,
            )
            out_dset.resize(len(in_dset), axis=0)
            start_idx = len(data)
        else:
            end_idx = start_idx + len(data)
            f_out[dset_info["name"]][start_idx:end_idx] = data
            start_idx = end_idx

        print("Memory peak: {0:.3f} MB".format(peak_memory_usage()))

    if start_idx != len(dset_info["dset"]):
        print(f"Warning: last index was {start_idx} not {len(dset_info['dset'])}")


def slicify(fancy_indices):
    """ [0,1,2, 6,7,8] --> [0:3, 6:9] """
    steps = np.diff(fancy_indices) != 1
    slice_starts = np.concatenate([fancy_indices[:1], fancy_indices[1:][steps]])
    slice_ends = np.concatenate([fancy_indices[:-1][steps], fancy_indices[-1:]]) + 1
    return [slice(slice_starts[i], slice_ends[i]) for i in range(len(slice_starts))]


def run_parser():
    # TODO deprecated
    warnings.warn("h5shuffle2 is deprecated and has been renamed to orcasong h5shuffle2")
    parser = argparse.ArgumentParser(
        description="Shuffle datasets in a h5file that have the same length. "
        "Uses chunkwise readout for speed-up."
    )
    parser.add_argument(
        "input_file", type=str, help="Path of the file that will be shuffled."
    )
    parser.add_argument(
        "--output_file",
        type=str,
        default=None,
        help="If given, this will be the name of the output file. "
        "Default: input_file + suffix.",
    )
    parser.add_argument(
        "--datasets",
        type=str,
        nargs="*",
        default=("x", "y"),
        help="Which datasets to include in output. Default: x, y",
    )
    parser.add_argument(
        "--max_ram_fraction",
        type=float,
        default=0.25,
        help="in [0, 1]. Fraction of all available ram to use for reading one batch of data "
        "Note: this should "
        "be <=~0.25 or so, since lots of ram is needed for in-memory shuffling. "
        "Default: 0.25",
    )
    parser.add_argument(
        "--iterations",
        type=int,
        default=None,
        help="Shuffle the file this many times. Default: Auto choose best number.",
    )
    parser.add_argument(
        "--max_ram",
        type=int,
        default=None,
        help="Available ram in bytes. Default: Use fraction of maximum "
             "available instead (see max_ram_fraction).",
    )
    h5shuffle2(**vars(parser.parse_args()))



from unittest import TestCase
import os
import orcasong
import orcasong.from_toml as from_toml

EXAMPLES = os.path.join(
    os.path.dirname(os.path.dirname(orcasong.__file__)), "examples"
)


def _test_extr(infile):
    return infile + "_extr"


orcasong.from_toml.EXTRACTORS["nu_chain_neutrino"] = _test_extr


class TestSetupProcessorExampleConfig(TestCase):
    def setUp(self):
        self.processor = from_toml.setup_processor(
            infile="test_in",
            toml_file=os.path.join(EXAMPLES, "orcasong_example.toml"),
            detx_file="test_det",
        )

    def test_time_window(self):
        self.assertEqual(self.processor.time_window, [-100, 5000])

    def test_max_n_hits(self):
        self.assertEqual(self.processor.max_n_hits, 2000)

    def test_chunksize(self):
        self.assertEqual(self.processor.chunksize, 16)

    def test_extractor_is_dummy_extractor(self):
        self.assertEqual(self.processor.extractor, "test_in_extr")



import os
from unittest import TestCase
import tempfile
import numpy as np
import h5py
import orcasong.core
import orcasong.extractors as extractors
from orcasong.plotting.plot_binstats import read_hists_from_h5file


__author__ = 'Stefan Reck'


test_dir = os.path.dirname(os.path.realpath(__file__))
MUPAGE_FILE = os.path.join(test_dir, "data", "mupage.root.h5")
DET_FILE = os.path.join(test_dir, "data", "KM3NeT_-00000001_20171212.detx")


class TestFileBinner(TestCase):
    """ Assert that the filebinner still produces the same output. """
    @classmethod
    def setUpClass(cls):
        cls.proc = orcasong.core.FileBinner(
            bin_edges_list=[
                ["pos_z", np.linspace(0, 200, 3)],
                ["time", np.linspace(0, 600, 3)],
                ["channel_id", np.linspace(-0.5, 30.5, 3)],
            ],
            extractor=extractors.get_real_data_info_extr(MUPAGE_FILE),
            det_file=DET_FILE,
            add_t0=True,
            keep_event_info=True,
        )
        cls.tmpdir = tempfile.TemporaryDirectory()
        cls.outfile = os.path.join(cls.tmpdir.name, "binned.h5")
        cls.proc.run(infile=MUPAGE_FILE, outfile=cls.outfile)
        cls.f = h5py.File(cls.outfile, "r")

    @classmethod
    def tearDownClass(cls):
        cls.f.close()
        cls.tmpdir.cleanup()

    def test_keys(self):
        self.assertSetEqual(set(self.f.keys()), {
            '_i_event_info', '_i_group_info', '_i_y', 'bin_stats',
            'event_info', 'group_info', 'x', 'x_indices', 'y'})

    def test_x(self):
        target = np.array([
            [[[4, 1], [6, 3]],
                [[12, 5], [6, 7]]],
            [[[4, 2], [1, 3]],
                [[5, 7], [8, 5]]],
            [[[3, 3], [2, 4]],
                [[5, 6], [6, 8]]]
        ], dtype=np.uint8)
        np.testing.assert_equal(target, self.f["x"])

    def test_y(self):
        y = self.f["y"][()]
        target = {
            'event_id': np.array([0., 1., 2.]),
            'run_id': np.array([1., 1., 1.]),
            'trigger_mask': np.array([18., 18., 16.]),
            'group_id': np.array([0, 1, 2]),
        }
        for k, v in target.items():
            np.testing.assert_equal(y[k], v)

    def test_bin_stats(self):
        bin_stats = read_hists_from_h5file(self.f)
        target_hists = {
            "channel_id": np.array([20.0, 8.0, 10.0, 8.0, 16.0, 13.0, 11.0, 9.0, 10.0, 11.0]),
            "pos_z": np.array([0.0, 0.0, 0.0, 36.0, 0.0, 19.0, 30.0, 0.0, 31.0, 0.0]),
            "time": np.array([57.0, 59.0]),
        }
        for dim, infos in bin_stats.items():
            np.testing.assert_equal(infos["hist"], target_hists[dim])


class TestFileGraph(TestCase):
    """ Assert that the FileGraph still produces the same output. """
    @classmethod
    def setUpClass(cls):
        cls.proc = orcasong.core.FileGraph(
            max_n_hits=3,
            time_window=[0, 50],
            hit_infos=["pos_z", "time", "channel_id"],
            extractor=extractors.get_real_data_info_extr(MUPAGE_FILE),
            det_file=DET_FILE,
            add_t0=False,
            keep_event_info=True,
            correct_timeslew=False,
        )
        cls.tmpdir = tempfile.TemporaryDirectory()
        cls.outfile = os.path.join(cls.tmpdir.name, "binned.h5")
        cls.proc.run(infile=MUPAGE_FILE, outfile=cls.outfile)
        cls.f = h5py.File(cls.outfile, "r")

    @classmethod
    def tearDownClass(cls):
        cls.f.close()
        cls.tmpdir.cleanup()

    def test_keys(self):
        self.assertSetEqual(set(self.f.keys()), {
            '_i_event_info', '_i_group_info', '_i_y',
            'event_info', 'group_info', 'x', 'x_indices', 'y'})

    def test_x_attrs(self):
        to_check = {
            "hit_info_0": "pos_z",
            "hit_info_1": "time",
            "hit_info_2": "channel_id",
            "hit_info_3": "is_valid",
        }
        attrs = dict(self.f["x"].attrs)
        for k, v in to_check.items():
            self.assertTrue(attrs[k] == v)

    def test_x(self):
        target = np.array([
            [[676.941,  13.,  30.,   1.],
             [461.111,  32.,   9.,   1.],
             [424.941,   1.,  30.,   1.]],
            [[172.83,  32.,  25.,   1.],
             [316.83,   2.,  14.,   1.],
             [461.059,   1.,   3.,   1.]],
            [[496.83,  34.,  25.,   1.],
             [605.111,   9.,   4.,   1.],
             [424.889,  46.,  29.,   1.]]
        ], dtype=np.float32)
        np.testing.assert_equal(target, self.f["x"])

    def test_y(self):
        y = self.f["y"][()]
        target = {
            'event_id': np.array([0., 1., 2.]),
            'run_id': np.array([1., 1., 1.]),
            'trigger_mask': np.array([18., 18., 16.]),
            'group_id': np.array([0, 1, 2]),
        }
        for k, v in target.items():
            np.testing.assert_equal(y[k], v)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Code that reads the h5 simulation files and extracts the necessary information for making event images."""

import numpy as np
import km3pipe as kp


def get_primary_track_index(event_blob):
    """
    Gets the index of the primary (neutrino) track.

    Uses bjorkeny in order to get the primary track, since bjorkeny!=0 for the initial interacting neutrino.

    Parameters
    ----------
    event_blob : kp.io.HDF5Pump.blob
        HDF5Pump event blob.

    Returns
    -------
    primary_index : int
        Index of the primary track (=neutrino) in the 'McTracks' branch.

    """
    bjorken_y_array = event_blob['McTracks'].bjorkeny
    primary_index = np.where(bjorken_y_array != 0.0)[0][0]
    return primary_index


def get_time_residual_nu_interaction_mean_triggered_hits(time_interaction, hits_time, triggered):
    """
    Gets the time_residual of the event with respect to mean time of the triggered hits.

    This is required for vertex_time reconstruction, as the absolute time scale needs to be relative to the triggered hits.

    Careful: sometimes, not the neutrino event is triggered, but just some random noise!
    This means that in very rare cases, the time_residual_vertex can be very large (Mio. of ns), which might throw off
    a NN with vertex_time reconstruction.

    Parameters
    ----------
    time_interaction : float
        Time of the neutrino interaction measured in JTE time.
    hits_time : ndarray(ndim=1)
        Time of the event_hits measured in JTE time.
    triggered : ndarray(ndim=1)
        Array with trigger flags that specifies if the hit is triggered or not.

    """
    hits_time_triggered = hits_time[triggered == 1]
    t_mean_triggered = np.mean(hits_time_triggered, dtype=np.float64)
    time_residual_vertex = t_mean_triggered - time_interaction

    return time_residual_vertex


def get_hits(event_blob, geo, do_mc_hits, data_cuts, do4d):
    """
    Returns a hits array that contains [pos_x, pos_y, pos_z, time, triggered, channel_id (optional)].

    Parameters
    ----------
    event_blob : kp.io.HDF5Pump.blob
        Event blob of the HDF5Pump which contains all information for one event.
    geo : kp.Geometry
        km3pipe Geometry instance that contains the geometry information of the detector.
        Only used if the event_blob is from a non-calibrated file!
    do_mc_hits : bool
        Tells the function of the hits (mc_hits + BG) or the mc_hits only should be parsed.
        In the case of mc_hits, the dom_id needs to be calculated thanks to the jpp output.
    data_cuts : dict
        Specifies if cuts should be applied.
        Contains the keys 'triggered' and 'energy_lower/upper_limit' and 'throw_away_prob'.
    do4d : tuple(bool, str)
        Tuple that declares if 4D histograms should be created [0] and if yes, what should be used as the 4th dim after xyz.
        In the case of 'channel_id', this information needs to be included in the event_hits as well.

    Returns
    -------
    event_hits : ndarray(ndim=2)
        2D array that contains the hits data for the input event [pos_x, pos_y, pos_z, time, triggered, (channel_id)].

    """
    # parse hits [x,y,z,time]
    hits = event_blob['Hits'] if do_mc_hits is False else event_blob['McHits']

    if 'pos_x' not in event_blob['Hits'].dtype.names: # check if blob already calibrated
        hits = geo.apply(hits)

    if data_cuts['triggered'] is True:
        hits = hits.triggered_hits # alternative, though it only works for the triggered condition!

    pos_x, pos_y, pos_z = hits.pos_x, hits.pos_y, hits.pos_z
    hits_time = hits.time
    triggered = hits.triggered

    ax = np.newaxis
    event_hits = np.concatenate([pos_x[:, ax], pos_y[:, ax], pos_z[:, ax], hits_time[:, ax], triggered[:, ax]], axis=1) # dtype: np.float64

    if do4d[0] is True and do4d[1] == 'channel_id' or do4d[1] == 'xzt-c':
        event_hits = np.concatenate([event_hits, hits.channel_id[:, ax]], axis=1)

    return event_hits


def get_tracks(event_blob, file_particle_type, event_hits, prod_ident):
    """
    Returns the event_track, which contains important event_info and mc_tracks data for the input event.

    Parameters
    ----------
    event_blob : kp.io.HDF5Pump.blob
        Event blob of the HDF5Pump which contains all information for one event.
    file_particle_type : str
        String that specifies the type of particles that are contained in the file: ['undefined', 'muon', 'neutrino'].
    event_hits : ndarray(ndim=2)
        2D array that contains the hits data for the input event.
    prod_ident : int
        Optional int that identifies the used production, more documentation in the docs of the main function.

    Returns
    -------
    event_track : ndarray(ndim=1)
        1D array that contains important event_info and mc_tracks data for the input event.

        If file_particle_type = 'undefined':
        [event_id, run_id, (prod_ident)].

        If file_particle_type = 'neutrino'/'muon':
        [event_id, particle_type, energy, is_cc, bjorkeny, dir_x, dir_y, dir_z, time_track, run_id,
        vertex_pos_x, vertex_pos_y, vertex_pos_z, time_residual_vertex/n_muons, (prod_ident)].

    """
    ## parse EventInfo and Header information

    # km3pipe event_id is the aanet frame_index
    # for random_noise files, multiple events have the same frame_index, so use the group_id instead
    if file_particle_type == 'undefined':
        event_id = event_blob['EventInfo'].group_id[0]
    else:
        event_id = event_blob['EventInfo'].event_id[0]

    if 'Header' in event_blob: # if Header exists in file, take run_id from it.
        run_id = event_blob['Header'].start_run.run_id.astype('float32')
    else:
        if file_particle_type == 'undefined': # currently used with random_noise files
            run_id = event_blob['EventInfo'].run_id
        else:
            raise ValueError('The run_id could not be read from the EventInfo or the Header, '
                             'please check the source code in get_tracks().')

    ## collect all event_track information, dependent on file_particle_type

    if file_particle_type == 'undefined':
        particle_type = 0
        frame_index = event_blob['EventInfo'].event_id[0]

        track = {'event_id': event_id, 'run_id': run_id, 'particle_type': particle_type, 'frame_index': frame_index}

    elif file_particle_type == 'muon':
        # take index 1, index 0 is the empty neutrino mc_track
        particle_type = event_blob['McTracks'][1].type # assumed that this is the same for all muons in a bundle
        is_cc = event_blob['McTracks'][1].is_cc # always 1 actually
        bjorkeny = event_blob['McTracks'][1].bjorkeny # always 0 actually
        time_interaction = event_blob['McTracks'][1].time  # same for all muons in a bundle
        n_muons = event_blob['McTracks'].shape[0] - 1 # takes position of time_residual_vertex in 'neutrino' case

        # sum up the energy of all muons
        energy = np.sum(event_blob['McTracks'].energy)

        # all muons in a bundle are parallel, so just take dir of first muon
        dir_x, dir_y, dir_z = event_blob['McTracks'][1].dir_x, event_blob['McTracks'][1].dir_y, event_blob['McTracks'][1].dir_z

        # vertex is the weighted (energy) mean of the individual vertices
        vertex_pos_x = np.average(event_blob['McTracks'][1:].pos_x, weights=event_blob['McTracks'][1:].energy)
        vertex_pos_y = np.average(event_blob['McTracks'][1:].pos_y, weights=event_blob['McTracks'][1:].energy)
        vertex_pos_z = np.average(event_blob['McTracks'][1:].pos_z, weights=event_blob['McTracks'][1:].energy)

        track = {'event_id': event_id, 'particle_type': particle_type, 'energy': energy, 'is_cc': is_cc,
                 'bjorkeny': bjorkeny, 'dir_x': dir_x, 'dir_y': dir_y, 'dir_z': dir_z,
                 'time_interaction': time_interaction,  'run_id': run_id, 'vertex_pos_x': vertex_pos_x,
                 'vertex_pos_y': vertex_pos_y, 'vertex_pos_z': vertex_pos_z, 'n_muons': n_muons}

    elif file_particle_type == 'neutrino':
        p = get_primary_track_index(event_blob)
        particle_type = event_blob['McTracks'][p].type
        energy = event_blob['McTracks'][p].energy
        is_cc = event_blob['McTracks'][p].is_cc
        bjorkeny = event_blob['McTracks'][p].bjorkeny
        dir_x, dir_y, dir_z = event_blob['McTracks'][p].dir_x, event_blob['McTracks'][p].dir_y, event_blob['McTracks'][p].dir_z
        time_interaction = event_blob['McTracks'][p].time  # actually always 0 for primary neutrino, measured in MC time
        vertex_pos_x, vertex_pos_y, vertex_pos_z = event_blob['McTracks'][p].pos_x, event_blob['McTracks'][p].pos_y, \
                                                   event_blob['McTracks'][p].pos_z

        hits_time, triggered = event_hits[:, 3], event_hits[:, 4]
        time_residual_vertex = get_time_residual_nu_interaction_mean_triggered_hits(time_interaction, hits_time, triggered)

        track = {'event_id': event_id, 'particle_type': particle_type, 'energy': energy, 'is_cc': is_cc,
                 'bjorkeny': bjorkeny, 'dir_x': dir_x, 'dir_y': dir_y, 'dir_z': dir_z,
                 'time_interaction': time_interaction,  'run_id': run_id, 'vertex_pos_x': vertex_pos_x,
                 'vertex_pos_y': vertex_pos_y, 'vertex_pos_z': vertex_pos_z,
                 'time_residual_vertex': time_residual_vertex}

    else:
        raise ValueError('The file_particle_type "', str(file_particle_type), '" is not known.')

    if prod_ident is not None: track['prod_ident'] = prod_ident

    dtypes = [(key, np.float64) for key in track.keys()]
    event_track = kp.dataclasses.Table(track, dtype=dtypes, h5loc='y', name='Event_Information')

    return event_track


class EventDataExtractor(kp.Module):
    """
    Class that takes a km3pipe blob which contains the information for one event and returns
    a blob with a hit array and a track array that contains all relevant information of the event.
    """
    def configure(self):
        """
        Sets up the input arguments of the EventDataExtractor class.
        """
        self.file_particle_type = self.require('file_particle_type')
        self.geo = self.require('geo')
        self.do_mc_hits = self.require('do_mc_hits')
        self.data_cuts = self.require('data_cuts')
        self.do4d = self.require('do4d')
        self.prod_ident = self.require('prod_ident')
        self.event_hits_key = self.get('event_hits', default='event_hits')
        self.event_track_key = self.get('event_track', default='event_track')

    def process(self, blob):
        """
        Returns a blob (dict), which contains the event_hits array and the event_track array.

        Parameters
        ----------
        blob : dict
            Km3pipe blob which contains all the data from the input file.

        Returns
        -------
        blob : dict
            Dictionary that contains the event_hits array and the event_track array.

        """
        blob[self.event_hits_key] = get_hits(blob, self.geo, self.do_mc_hits, self.data_cuts, self.do4d)
        blob[self.event_track_key] = get_tracks(blob, self.file_particle_type, blob[self.event_hits_key], self.prod_ident)
        return blob





import os
import time
import h5py
import numpy as np
import argparse
import warnings


__author__ = 'Stefan Reck, Michael Moser'


class FileConcatenator:
    """
    For concatenating many small h5 files to a single large one in
    km3pipe-compatible format.

    Parameters
    ----------
    input_files : List
        List that contains all filepaths of the input files.
    skip_errors : bool
        If true, ignore files that can't be concatenated.
    comptopts_update : dict, optional
        Overwrite the compression options that get read from the
        first file. E.g. {'chunksize': 10} to get a chunksize of 10.

    Attributes
    ----------
    comptopts : dict
        Options for compression. They are read from the first input file,
        but they can be updated as well during init.
    cumu_rows : np.array
        The cumulative number of rows (axis_0) of the specified
        input .h5 files (i.e. [0,100,200,300,...] if each file has 100 rows).

    """
    def __init__(self, input_files, skip_errors=False, comptopts_update=None):
        self.skip_errors = skip_errors
        print(f"Checking {len(input_files)} files ...")

        self.input_files, self.cumu_rows = self._get_cumu_rows(input_files)
        print(f"Total rows:\t{self.cumu_rows[-1]}")

        # Get compression options from first file in the list
        self.comptopts = get_compopts(self.input_files[0])
        if comptopts_update:
            self.comptopts.update(comptopts_update)
        print("\n".join([f"  {k}:\t{v}" for k, v in self.comptopts.items()]))

        self._modify_folder = False

    @classmethod
    def from_list(cls, list_file, n_files=None, **kwargs):
        """
        Initialize with a .txt list containing the filepaths.

        Parameters
        ----------
        list_file : str
            Path to a txt file containing the input filepaths, one per line.
        n_files : int, optional
            Only load these many files from the list.

        """
        input_files = []
        with open(list_file) as f:
            for line in f:
                filepath = line.rstrip('\n')
                if filepath != "":
                    input_files.append(filepath)
        if n_files is not None:
            input_files = input_files[:n_files]
        return cls(input_files, **kwargs)

    def concatenate(self, output_filepath, append_used_files=True):
        """
        Concatenate the input files.

        Parameters
        ----------
        output_filepath : str
            Path of the concatenated output file.
        append_used_files : bool
            If True (default), add a dataset called 'used_files' to the
            output that contains the paths of the input_files.

        """
        print(f"Creating file {output_filepath}")
        with h5py.File(output_filepath, 'x') as f_out:
            start_time = time.time()
            for input_file_nmbr, input_file in enumerate(self.input_files):
                print(f'Processing file {input_file_nmbr+1}/'
                      f'{len(self.input_files)}: {input_file}')
                with h5py.File(input_file, 'r') as f_in:
                    self._conc_file(f_in, f_out, input_file, input_file_nmbr)
                f_out.flush()
            elapsed_time = time.time() - start_time

            if append_used_files:
                print("Adding used files to output")
                f_out.create_dataset(
                    "used_files",
                    data=[n.encode("ascii", "ignore") for n in self.input_files]
                )

        copy_attrs(self.input_files[0], output_filepath)

        print(f"\nConcatenation complete!"
              f"\nElapsed time: {elapsed_time/60:.2f} min "
              f"({elapsed_time/len(self.input_files):.2f} s per file)")

    def _conc_file(self, f_in, f_out, input_file, input_file_nmbr):
        """ Conc one file to the output. """
        for folder_name in f_in:
            if is_folder_ignored(folder_name):
                # we dont need datasets created by pytables anymore
                continue
            input_dataset = f_in[folder_name]
            folder_data = input_dataset[()]

            if input_file_nmbr > 0:
                # we need to add the current number of the
                # group_id / index in the file_output to the
                # group_ids / indices of the file that is to be appended
                try:
                    if folder_name.endswith("_indices") and \
                            "index" in folder_data.dtype.names:
                        column_name = "index"
                    elif "group_id" in folder_data.dtype.names:
                        column_name = "group_id"
                    else:
                        column_name = None
                except TypeError:
                    column_name = None
                if column_name is not None:
                    # add 1 because the group_ids / indices start with 0
                    folder_data[column_name] += \
                        np.amax(f_out[folder_name][column_name]) + 1

            if self._modify_folder:
                data_mody = self._modify(
                    input_file, folder_data, folder_name)
                if data_mody is not None:
                    folder_data = data_mody

            if input_file_nmbr == 0:
                # first file; create the dataset
                dset_shape = (self.cumu_rows[-1],) + folder_data.shape[1:]
                print(f"\tCreating dataset '{folder_name}' with shape {dset_shape}")
                output_dataset = f_out.create_dataset(
                    folder_name,
                    data=folder_data,
                    maxshape=dset_shape,
                    chunks=(self.comptopts["chunksize"],) + folder_data.shape[1:],
                    compression=self.comptopts["complib"],
                    compression_opts=self.comptopts["complevel"],
                    shuffle=self.comptopts["shuffle"],
                )
                output_dataset.resize(self.cumu_rows[-1], axis=0)

            else:
                f_out[folder_name][
                    self.cumu_rows[input_file_nmbr]:self.cumu_rows[input_file_nmbr + 1]] = folder_data

    def _modify(self, input_file, folder_data, folder_name):
        raise NotImplementedError

    def _get_cumu_rows(self, input_files):
        """
        Get the cumulative number of rows of the input_files.
        Also checks if all the files can be safely concatenated to the
        first one.

        """
        # names of datasets that will be in the output; read from first file
        with h5py.File(input_files[0], 'r') as f:
            keys_stripped = strip_keys(list(f.keys()))

        errors, rows_per_file, valid_input_files = [], [0], []
        for i, file_name in enumerate(input_files, start=1):
            file_name = os.path.abspath(file_name)
            try:
                rows_this_file = _get_rows(file_name, keys_stripped)
            except Exception as e:
                errors.append(e)
                warnings.warn(f"Error during check of file {i}: {file_name}")
                continue
            valid_input_files.append(file_name)
            rows_per_file.append(rows_this_file)

        if errors:
            print("\n------- Errors -------\n----------------------")
            for error in errors:
                warnings.warn(str(error))
                print("")
            err_str = f"{len(errors)} error(s) during check of files! See above"
            if self.skip_errors:
                warnings.warn(err_str)
            else:
                raise OSError(err_str)

        print(f"Valid input files: {len(valid_input_files)}/{len(input_files)}")
        print("Datasets:\t" + ", ".join(keys_stripped))
        return valid_input_files, np.cumsum(rows_per_file)


def _get_rows(file_name, target_datasets):
    """ Get no of rows from a file and check if its good for conc'ing. """
    with h5py.File(file_name, 'r') as f:
        # check if all target datasets are in the file
        if not all(k in f.keys() for k in target_datasets):
            raise KeyError(
                f"File {file_name} does not have the "
                f"keys of the first file! "
                f"It has {f.keys()} First file: {target_datasets}"
            )
        # check if all target datasets in the file have the same length
        rows = [f[k].shape[0] for k in target_datasets]
        if not all(row == rows[0] for row in rows):
            raise ValueError(
                f"Datasets in file {file_name} have varying length! "
                f"{dict(zip(target_datasets, rows))}"
            )
        # check if the file has additional datasets apart from the target keys
        if not all(k in target_datasets for k in strip_keys(list(f.keys()))):
            warnings.warn(
                f"Additional datasets found in file {file_name} compared "
                f"to the first file, they wont be in the output! "
                f"This file: {strip_keys(list(f.keys()))} "
                f"First file {target_datasets}"
            )
    return rows[0]


def strip_keys(f_keys):
    """ Remove unwanted keys from list. """
    return [x for x in f_keys if not is_folder_ignored(x)]


def is_folder_ignored(folder_name):
    """
    Defines datasets which should be ignored during concat.

    Remove pytables folders starting with '_i_', because the shape
    of its first axis does not correspond to the number of events
    in the file. All other folders normally have an axis_0 shape
    that is equal to the number of events in the file.
    Also remove bin_stats.

    """
    return any([s in folder_name for s in (
        '_i_', "bin_stats", "raw_header", "header")])


def get_compopts(file):
    """
    Get the following compression options from a h5 file as a dict:

    complib : str
        Specifies the compression library that should be used for saving
        the concatenated output files.
    complevel : None/int
        Specifies the compression level that should be used for saving
        the concatenated output files.
        A compression level is only available for gzip compression, not lzf!
    chunksize : None/int
        Specifies the chunksize for axis_0 in the concatenated output files.
    shuffle : bool
        Enable shuffle filter for chunks.

    """
    with h5py.File(file, 'r') as f:
        dset = f[strip_keys(list(f.keys()))[0]]
        comptopts = {}
        comptopts["complib"] = dset.compression
        if comptopts["complib"] == 'lzf':
            comptopts["complevel"] = None
        else:
            comptopts["complevel"] = dset.compression_opts
        comptopts["chunksize"] = dset.chunks[0]
        comptopts["shuffle"] = dset.shuffle
    return comptopts


def copy_attrs(source_file, target_file):
    """
    Copy file and dataset attributes from one h5 file to another.
    """
    print("Copying attributes")
    with h5py.File(source_file, "r") as src:
        with h5py.File(target_file, "a") as trg:
            _copy_attrs(src, trg)
            for dset_name, target_dataset in trg.items():
                if dset_name in src:
                    _copy_attrs(src[dset_name], target_dataset)


def _copy_attrs(src_datset, target_dataset):
    for k in src_datset.attrs.keys():
        try:
            if k not in target_dataset.attrs:
                target_dataset.attrs[k] = src_datset.attrs[k]
        except TypeError as e:
            # above can fail if attr is bool and created using pt
            warnings.warn(f"Error: Can not copy attribute {k}: {e}")


def concatenate(file, outfile="concatenated.h5", no_used_files=False, skip_errors=False):
    """ Concatenate wrapped in a function. """
    if len(file) == 1:
        fc = FileConcatenator.from_list(
            file[0],
            skip_errors=skip_errors
        )
    else:
        fc = FileConcatenator(
            input_files=file,
            skip_errors=skip_errors
        )
    fc.concatenate(
        outfile,
        append_used_files=not no_used_files,
    )


def main():
    # TODO deprecated
    warnings.warn("concatenate is deprecated and has been renamed to orcasong concatenate")
    parser = argparse.ArgumentParser(
        description='Concatenate many small h5 files to a single large one '
                    'in a km3pipe compatible format. This is intended for '
                    'files that get generated by orcasong, i.e. all datsets '
                    'should have the same length, with one row per '
                    'blob. '
                    'Compression options and the datasets to be created in '
                    'the new file will be read from the first input file.')
    parser.add_argument(
        'file', type=str, nargs="*",
        help="Define the files to concatenate. If it's one argument: A txt list "
             "with pathes of h5 files to concatenate (one path per line). "
             "If it's multiple arguments: "
             "The pathes of h5 files to concatenate.")
    parser.add_argument(
        '--outfile', type=str, default="concatenated.h5",
        help='The absoulte filepath of the output .h5 file that will be created. ')
    parser.add_argument(
        '--no_used_files', action='store_true',
        help="Per default, the paths of the input files are added "
             "as their own datagroup in the output file. Use this flag to "
             "disable. ")
    parser.add_argument(
        '--skip_errors', action='store_true',
        help="If true, ignore files that can't be concatenated. ")
    concatenate(**vars(parser.parse_args()))


if __name__ == '__main__':
    main()



"""
Run with a parser to plot the binning statistics.
Functions for plotting the bin stats made by the BinningStatsMaker module.
"""

import os
import warnings
import argparse
import matplotlib
matplotlib.use("pdf")
import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages
import h5py
import numpy as np


__author__ = 'Stefan Reck'


def plot_hists(hists_list, save_to, plot_bin_edges=True):
    """
    Plot histograms made by the BinningStatsMaker to the given pdf path.

    Parameters
    ----------
    hists_list : List
        Can also be a list of dicts (for multiple files), which will be
        combined to one plot.
        Each dict has info about the binning, generated by the BinningStatsMaker.
    save_to : str
        Where to save the plot to.
    plot_bin_edges : bool
        If true, will plot the bin edges as horizontal lines. Is never used
        for the time binning (field name "time").

    """
    hists = combine_hists(hists_list)

    if os.path.exists(save_to):
        raise FileExistsError(f"File {save_to} exists already!")

    with PdfPages(save_to) as pdf_file:
        for bin_name, hists_data in hists.items():
            hist_bin_edges = hists_data["hist_bin_edges"]
            bin_edges = hists_data["bin_edges"]
            hist = hists_data["hist"]
            cut_off = hists_data["cut_off"]
            total_hits = np.sum(hist) + np.sum(cut_off)

            bin_widths = np.diff(hist_bin_edges)
            hist_prob = hist / bin_widths / np.sum(hist)

            fig, ax = plt.subplots()
            plt.bar(
                hist_bin_edges[:-1],
                hist_prob,
                align="edge",
                width=bin_widths
            )

            if plot_bin_edges and bin_name != "time":
                for bin_edge in bin_edges:
                    plt.axvline(
                        x=bin_edge, color='grey', linestyle='-',
                        linewidth=1, alpha=0.9
                    )

            # place a text box in upper left in axes coords
            out_neg_rel = cut_off[0] / total_hits
            out_pos_rel = cut_off[1] / total_hits
            textstr = "Hits cut off:\nLeft: {:.1%}\nRight: {:.1%}".format(
                out_neg_rel, out_pos_rel)
            props = dict(boxstyle='round', facecolor='white', alpha=0.9)
            ax.text(
                0.95, 0.95, textstr, transform=ax.transAxes,
                verticalalignment='top', bbox=props,
                horizontalalignment="right"
            )

            # the auto ticks are nice even numbers
            ylims = [0, ax.get_yticks().max()]
            ax.set_ylim(ylims)

            plt.xlabel(bin_name)
            plt.ylabel("Density of hits")

            pdf_file.savefig(fig)

    print("Saved binning plot to " + save_to)


def combine_hists(hists_list):
    """
    Combine the hists for multiple files into a single big one.

    Parameters
    ----------
    hists_list : List
        A list of dicts (hists from the BinningStatsMaker for different files).

    Returns
    -------
    combined_hists : Dict
        The combined hist.

    """
    if len(hists_list) == 1:
        # just one entry, no need to combine
        return hists_list[0]

    # initialize combined dict according to first dict in list
    combined_hists = {}
    for bin_name, hists_data in hists_list[0].items():
        hist_bin_edges = hists_data["hist_bin_edges"]
        bin_edges = hists_data["bin_edges"]
        bin_name = bin_name

        combined_hists[bin_name] = {
            "hist": np.zeros(len(hist_bin_edges) - 1),
            "hist_bin_edges": hist_bin_edges,
            "bin_edges": bin_edges,
            # below smallest edge, above largest edge:
            "cut_off": np.zeros(2),
        }

    # add them together
    for hists in hists_list:
        for bin_name, hists_data in hists.items():
            # name and bin edges must be the same for all hists
            if bin_name not in combined_hists:
                raise NameError(
                    "Hists dont all have the same binning field name:"
                    "{}".format(bin_name))

            bin_edges = hists_data["bin_edges"]
            comb_edges = combined_hists[bin_name]["bin_edges"]
            if len(bin_edges) != len(comb_edges):
                raise ValueError("Hists have different hist bin edges: {} vs {}"
                                 .format(bin_edges, comb_edges))

            hist_bin_edges = hists_data["hist_bin_edges"]
            comb_hist_edges = combined_hists[bin_name]["hist_bin_edges"]
            if len(hist_bin_edges) != len(comb_hist_edges):
                raise ValueError("Hists have different bin edges: {} vs {}"
                                 .format(hist_bin_edges, comb_hist_edges))

            combined_hists[bin_name]["hist"] += hists_data["hist"]
            combined_hists[bin_name]["cut_off"] += hists_data["cut_off"]

    return combined_hists


def add_hists_to_h5file(hists, f):
    """
    Add the binning stats as groups to a h5 file.

    Parameters
    ----------
    hists : dict
        The histst from the BinningStatsMaker module.
    f
        The opened h5 file.

    """
    for bin_name, hists_data in hists.items():
        for key, val in hists_data.items():
            f.create_dataset(f"bin_stats/{bin_name}/{key}", data=val)


def read_hists_from_h5file(f):
    """ Read the binning stats from the file, put it in dicts. """
    data = {}
    for bin_name, hists_data in f["bin_stats/"].items():
        data[bin_name] = {k: v[()] for k, v in hists_data.items()}
    return data


def plot_hist_of_files(save_as, files=None):
    """
    Plot the binning stats of all h5 files in the current directory.

    Parameters
    ----------
    save_as : str
        Where to save the plot to.
    files : List, optional
        Path of files to use instead.

    """
    if not files:
        files = get_all_h5_files()

    hists_list = []
    print("Plotting stats of {} file(s)".format(len(files)))

    for i, file in enumerate(files, start=1):
        if i % 100 == 0:
            print(f"Reading file {i} / {len(files)} ...")
        try:
            with h5py.File(file, "r") as f:
                hists_list.append(read_hists_from_h5file(f))
        except KeyError:
            warnings.warn(f"ERROR: No bin_stats in {file}. Skipping ...")
            continue
    if len(hists_list) == 0:
        raise ValueError("No files with bin_stats found!")
    print("Plotting...")
    plot_hists(hists_list, save_as)


def get_all_h5_files():
    """ Get a list of all h5 files in the cwd. """
    files = []
    for file in os.listdir(os.getcwd()):
        if file.endswith(".h5"):
            files.append(file)
    return files


def main():
    # TODO deprecated
    warnings.warn(
        "plot_binstats is deprecated and has been renamed to orcasong plot_binstats")
    parser = argparse.ArgumentParser(
        description='Generate a plot with statistics of the binning. '
                    'Can only be used on files generated with the FileBinner when '
                    'add_bin_stats was set to true (default). ')
    parser.add_argument(
        '--save_as', type=str, default="bin_stats_plot.pdf",
        help='Filename of the plot. Default: bin_stats_plot.pdf.')
    parser.add_argument(
        'files', type=str, nargs='*', default=None,
        help='File(s) to plot. Default: Plot for all h5 files in current dir.')

    plot_hist_of_files(**vars(parser.parse_args()))


def add_parser(subparsers):
    parser = subparsers.add_parser(
        "plot_binstats",
        description='Generate a plot with statistics of the binning. '
                    'Can only be used on files generated with the FileBinner when '
                    'add_bin_stats was set to true (default). ')
    parser.add_argument(
        '--save_as', type=str, default="bin_stats_plot.pdf",
        help='Filename of the plot. Default: bin_stats_plot.pdf.')
    parser.add_argument(
        'files', type=str, nargs='*', default=None,
        help='File(s) to plot. Default: Plot for all h5 files in current dir.')
    parser.set_defaults(func=plot_hist_of_files)


if __name__ == "__main__":
    main()



import toml
import orcasong.core
import orcasong.extractors as extractors

# built-in extractors. First argument has to be the input filename,
# other parameters can be set via 'extractor_config' dict in the toml
EXTRACTORS = {
    "nu_chain_neutrino": extractors.get_neutrino_mc_info_extr,
    "nu_chain_muon": extractors.get_muon_mc_info_extr,
    "nu_chain_noise": extractors.get_random_noise_mc_info_extr,
    "nu_chain_data": extractors.get_real_data_info_extr,
    # TODO "bundle_mc": ???,
    # TODO "bundle_data": ???,
}

MODES = {
    "graph": orcasong.core.FileGraph,
    "image": orcasong.core.FileBinner,
}


def add_parser_run(subparsers):
    parser = subparsers.add_parser(
        "run",
        description='Produce a dl file from an aanet file.')
    parser.add_argument('infile', type=str, help="Aanet file in h5 format.")
    parser.add_argument('toml_file', type=str, help="Orcasong configuration in toml format.")
    parser.add_argument('--detx_file', type=str, default=None, help=(
        "Optional detx file to calibrate on the fly. Can not be used if a "
        "detx_file has also been given in the toml file."))
    parser.add_argument('--outfile', type=str, default=None, help=(
        "Path to output file. Default: Save with auto generated name in cwd."))
    parser.set_defaults(func=run_orcasong)


def run_orcasong(infile, toml_file, detx_file=None, outfile=None):
    setup_processor(infile, toml_file, detx_file).run(
        infile=infile, outfile=outfile)


def setup_processor(infile, toml_file, detx_file=None):
    cfg = toml.load(toml_file)
    processor = _get_verbose(cfg.pop("mode"), MODES)

    if "detx_file" in cfg:
        if detx_file is not None:
            raise ValueError("detx_file passed to run AND defined in toml")
        detx_file = cfg.pop("detx_file")

    if "extractor" in cfg:
        extractor_cfg = cfg.pop("extractor_config", {})
        extractor = _get_verbose(cfg.pop("extractor"), EXTRACTORS)(infile, **extractor_cfg)
    else:
        extractor = None

    return processor(
        det_file=detx_file,
        extractor=extractor,
        **cfg,
    )


def _get_verbose(key, d):
    if key not in d:
        raise KeyError(f"Unknown key '{key}' (available: {list(d.keys())})")
    return d[key]
